#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import pytest


test_api_traces = [
    "hsa_api_traces",
    "marker_api_traces",
    "hip_api_traces",
    "rccl_api_traces",
    "scratch_memory_traces",
]


# helper function
def node_exists(name, data, min_len=1):
    assert name in data
    assert data[name] is not None
    if isinstance(data[name], (list, tuple, dict, set)):
        assert len(data[name]) >= min_len, f"{name}:\n{data}"


def get_operation(record, kind_name, op_name=None):
    for idx, itr in enumerate(record["names"]):
        if kind_name == itr["kind"]:
            if op_name is None:
                return idx, itr["operations"]
            else:
                for oidx, oname in enumerate(itr["operations"]):
                    if op_name == oname:
                        return oidx

    return None


def get_operation_name(record, kind_idx, op_idx):
    for idx, itr in enumerate(record["names"]):
        if idx == kind_idx:
            return itr["operations"][op_idx]

    return None


def groupby_corr_id(trace_item, op_id=None):
    """
    If op_id is not none, returns records only with that operation ID

    {
        corr_id-1: record with internal corr_id = corr_id-1
        corr_id-2: record with internal corr_id = corr_id-2
        ...
    }
    """
    from rocprofiler_sdk.pytest_utils.dotdict import dotdict

    ret = {}

    for x in trace_item:
        if op_id is not None and x.operation != op_id:
            continue

        corr_id = x.correlation_id["internal"]

        if corr_id in ret.keys():
            assert False, f"Duplicate internal corr_id {corr_id}"
        else:
            ret[corr_id] = x

    return dotdict(ret)


def test_data_structure(input_data):
    """verify minimum amount of expected data is present"""
    data = input_data

    node_exists("rocprofiler-sdk-json-tool", data)

    sdk_data = data["rocprofiler-sdk-json-tool"]

    node_exists("metadata", sdk_data)
    node_exists("pid", sdk_data["metadata"])
    node_exists("main_tid", sdk_data["metadata"])
    node_exists("init_time", sdk_data["metadata"])
    node_exists("fini_time", sdk_data["metadata"])

    node_exists("agents", sdk_data)
    node_exists("call_stack", sdk_data)
    node_exists("callback_records", sdk_data)
    node_exists("buffer_records", sdk_data)

    node_exists("names", sdk_data["callback_records"])
    node_exists("code_objects", sdk_data["callback_records"])
    node_exists("kernel_symbols", sdk_data["callback_records"])
    node_exists("host_functions", sdk_data["callback_records"])
    node_exists("hsa_api_traces", sdk_data["callback_records"])
    node_exists("hip_api_traces", sdk_data["callback_records"], 0)
    node_exists("marker_api_traces", sdk_data["callback_records"])
    node_exists("kernel_dispatch", sdk_data["callback_records"])
    node_exists("memory_copies", sdk_data["callback_records"], 24)

    node_exists("names", sdk_data["buffer_records"])
    node_exists("kernel_dispatch", sdk_data["buffer_records"])
    node_exists("memory_copies", sdk_data["buffer_records"], 12)
    node_exists("hsa_api_traces", sdk_data["buffer_records"])
    node_exists("hip_api_traces", sdk_data["buffer_records"], 0)
    node_exists("marker_api_traces", sdk_data["buffer_records"])
    node_exists("retired_correlation_ids", sdk_data["buffer_records"])


def test_size_entries(input_data):
    # check that size fields are > 0 but account for function arguments
    # which are named "size"
    def check_size(data, bt):
        if "size" in data.keys():
            if isinstance(data["size"], str) and bt.endswith('["args"]'):
                pass
            else:
                assert data["size"] > 0, f"origin: {bt}"

    # recursively check the entire data structure
    def iterate_data(data, bt):
        if isinstance(data, (list, tuple)):
            for i, itr in enumerate(data):
                if isinstance(itr, dict):
                    check_size(itr, f"{bt}[{i}]")
                iterate_data(itr, f"{bt}[{i}]")
        elif isinstance(data, dict):
            check_size(data, f"{bt}")
            for key, itr in data.items():
                iterate_data(itr, f'{bt}["{key}"]')

    # start recursive check over entire JSON dict
    iterate_data(input_data, "input_data")


def test_timestamps(input_data):
    data = input_data
    sdk_data = data["rocprofiler-sdk-json-tool"]

    cb_start = {}
    cb_end = {}
    for titr in test_api_traces:
        for itr in sdk_data["callback_records"][titr]:
            cid = itr["correlation_id"]["internal"]
            phase = itr["phase"]
            if phase == 1:
                cb_start[cid] = itr["timestamp"]
            elif phase == 2:
                cb_end[cid] = itr["timestamp"]
                assert cb_start[cid] <= itr["timestamp"]
            else:
                assert phase == 1 or phase == 2

        for itr in sdk_data["buffer_records"][titr]:
            assert itr["start_timestamp"] <= itr["end_timestamp"]

    for titr in ["kernel_dispatch", "memory_copies"]:
        for itr in sdk_data["buffer_records"][titr]:
            assert itr["start_timestamp"] < itr["end_timestamp"], f"[{titr}] {itr}"
            assert itr["correlation_id"]["internal"] > 0, f"[{titr}] {itr}"
            assert itr["correlation_id"]["external"] > 0, f"[{titr}] {itr}"
            assert (
                sdk_data["metadata"]["init_time"] < itr["start_timestamp"]
            ), f"[{titr}] {itr}"
            assert (
                sdk_data["metadata"]["init_time"] < itr["end_timestamp"]
            ), f"[{titr}] {itr}"
            assert (
                sdk_data["metadata"]["fini_time"] > itr["start_timestamp"]
            ), f"[{titr}] {itr}"
            assert (
                sdk_data["metadata"]["fini_time"] > itr["end_timestamp"]
            ), f"[{titr}] {itr}"

            api_start = cb_start[itr["correlation_id"]["internal"]]
            # api_end = cb_end[itr["correlation_id"]["internal"]]
            assert api_start < itr["start_timestamp"], f"[{titr}] {itr}"
            # assert api_end <= itr["end_timestamp"], f"[{titr}] {itr}"


def test_internal_correlation_ids(input_data):
    data = input_data
    sdk_data = data["rocprofiler-sdk-json-tool"]

    api_corr_ids = []
    for titr in test_api_traces:
        for itr in sdk_data["callback_records"][titr]:
            api_corr_ids.append(itr["correlation_id"]["internal"])

        for itr in sdk_data["buffer_records"][titr]:
            api_corr_ids.append(itr["correlation_id"]["internal"])

    api_corr_ids_sorted = sorted(api_corr_ids)
    api_corr_ids_unique = list(set(api_corr_ids))

    for itr in sdk_data["buffer_records"]["kernel_dispatch"]:
        assert itr["correlation_id"]["internal"] in api_corr_ids_unique

    for itr in sdk_data["buffer_records"]["memory_copies"]:
        assert itr["correlation_id"]["internal"] in api_corr_ids_unique

    len_corr_id_unq = len(api_corr_ids_unique)
    assert len(api_corr_ids) != len_corr_id_unq
    assert max(api_corr_ids_sorted) == len_corr_id_unq


def test_external_correlation_ids(input_data):
    data = input_data
    sdk_data = data["rocprofiler-sdk-json-tool"]

    extern_corr_ids = []
    for titr in test_api_traces:
        for itr in sdk_data["callback_records"][titr]:
            assert itr["correlation_id"]["external"] > 0
            assert itr["thread_id"] == itr["correlation_id"]["external"]
            extern_corr_ids.append(itr["correlation_id"]["external"])

    extern_corr_ids = list(set(sorted(extern_corr_ids)))
    for titr in test_api_traces:
        for itr in sdk_data["buffer_records"][titr]:
            assert itr["correlation_id"]["external"] > 0, f"[{titr}] {itr}"
            assert (
                itr["thread_id"] == itr["correlation_id"]["external"]
            ), f"[{titr}] {itr}"
            assert itr["thread_id"] in extern_corr_ids, f"[{titr}] {itr}"
            assert itr["correlation_id"]["external"] in extern_corr_ids, f"[{titr}] {itr}"

    for titr in ["kernel_dispatch", "memory_copies"]:
        for itr in sdk_data["buffer_records"][titr]:
            assert itr["correlation_id"]["external"] > 0, f"[{titr}] {itr}"
            assert itr["correlation_id"]["external"] in extern_corr_ids, f"[{titr}] {itr}"

        for itr in sdk_data["callback_records"][titr]:
            assert itr["correlation_id"]["external"] > 0, f"[{titr}] {itr}"
            assert itr["correlation_id"]["external"] in extern_corr_ids, f"[{titr}] {itr}"


def test_kernel_ids(input_data):
    data = input_data
    sdk_data = data["rocprofiler-sdk-json-tool"]

    symbol_info = {}
    for itr in sdk_data["callback_records"]["kernel_symbols"]:
        phase = itr["phase"]
        payload = itr["payload"]
        kern_id = payload["kernel_id"]

        assert phase == 1 or phase == 2
        assert kern_id > 0
        if phase == 1:
            assert len(payload["kernel_name"]) > 0
            symbol_info[kern_id] = payload
        elif phase == 2:
            assert payload["kernel_id"] in symbol_info.keys()
            assert payload["kernel_name"] == symbol_info[kern_id]["kernel_name"]

    for itr in sdk_data["buffer_records"]["kernel_dispatch"]:
        assert itr["dispatch_info"]["kernel_id"] in symbol_info.keys()

    for itr in sdk_data["callback_records"]["kernel_dispatch"]:
        assert itr["payload"]["dispatch_info"]["kernel_id"] in symbol_info.keys()


def test_kernel_dispatch_ids(input_data):
    data = input_data
    sdk_data = data["rocprofiler-sdk-json-tool"]

    num_dispatches = len(sdk_data["buffer_records"]["kernel_dispatch"])
    num_cb_dispatches = len(sdk_data["callback_records"]["kernel_dispatch"])

    assert num_cb_dispatches == (3 * num_dispatches)

    bf_seq_ids = []
    for itr in sdk_data["buffer_records"]["kernel_dispatch"]:
        bf_seq_ids.append(itr["dispatch_info"]["dispatch_id"])

    cb_seq_ids = []
    for itr in sdk_data["callback_records"]["kernel_dispatch"]:
        cb_seq_ids.append(itr["payload"]["dispatch_info"]["dispatch_id"])

    bf_seq_ids = sorted(bf_seq_ids)
    cb_seq_ids = sorted(cb_seq_ids)

    assert (3 * len(bf_seq_ids)) == len(cb_seq_ids)

    assert bf_seq_ids[0] == cb_seq_ids[0]
    assert bf_seq_ids[-1] == cb_seq_ids[-1]

    def get_uniq(data):
        return list(set(data))

    bf_seq_ids_uniq = get_uniq(bf_seq_ids)
    cb_seq_ids_uniq = get_uniq(cb_seq_ids)

    assert bf_seq_ids == bf_seq_ids_uniq
    assert len(cb_seq_ids) == (3 * len(cb_seq_ids_uniq))
    assert len(bf_seq_ids) == num_dispatches
    assert len(bf_seq_ids_uniq) == num_dispatches
    assert len(cb_seq_ids_uniq) == num_dispatches


def test_async_copy_direction(input_data):
    data = input_data
    sdk_data = data["rocprofiler-sdk-json-tool"]

    # Direction values:
    #   0 == ??? (unknown)
    #   1 == H2H (host to host)
    #   2 == H2D (host to device)
    #   3 == D2H (device to host)
    #   4 == D2D (device to device)
    default_async_dir_cnt = dict([(idx, 0) for idx in range(0, 5)])
    thread_async_dir_cnt = {}
    for itr in sdk_data.buffer_records.memory_copies:
        tid = itr.thread_id
        if tid not in thread_async_dir_cnt.keys():
            thread_async_dir_cnt[tid] = default_async_dir_cnt
        op_id = itr.operation
        assert op_id > 1, f"{itr}"
        assert op_id < 4, f"{itr}"
        thread_async_dir_cnt[tid][op_id] += 1

    for itr in sdk_data.callback_records.memory_copies:
        tid = itr.thread_id
        if tid not in thread_async_dir_cnt.keys():
            thread_async_dir_cnt[tid] = default_async_dir_cnt
        op_id = itr.operation
        assert op_id > 1, f"{itr}"
        assert op_id < 4, f"{itr}"
        thread_async_dir_cnt[tid][op_id] += 1

        phase = itr.phase
        pitr = itr.payload

        assert phase is not None, f"{itr}"
        assert pitr is not None, f"{itr}"

        if phase == 1:
            assert pitr.start_timestamp == 0, f"{itr}"
            assert pitr.end_timestamp == 0, f"{itr}"
        elif phase == 2:
            assert pitr.start_timestamp > 0, f"{itr}"
            assert pitr.end_timestamp > 0, f"{itr}"
            assert pitr.end_timestamp >= pitr.start_timestamp, f"{itr}"
        else:
            assert phase == 1 or phase == 2, f"{itr}"

    # in the transpose test which generates the input file,
    # two threads each perform one H2D + one D2H memory copy.
    # there are at least two callback records (phase start +
    # phase end) and one buffer record for each memory copy,
    # i.e., at least 3 records per memory copy
    assert len(thread_async_dir_cnt) == 2, f"{thread_async_dir_cnt}"
    for tid, async_dir_cnt in thread_async_dir_cnt.items():
        min_copy_records = 3
        assert async_dir_cnt[0] == 0
        assert async_dir_cnt[1] == 0
        assert async_dir_cnt[2] >= min_copy_records, f"TID={tid}:\n\t{async_dir_cnt}"
        assert async_dir_cnt[3] >= min_copy_records, f"TID={tid}:\n\t{async_dir_cnt}"
        assert async_dir_cnt[4] == 0
        # HIP memory copies may be decomposed into more than one
        # memory copy at the HSA level so require it to be a multiple
        # of min_copy_records
        assert (
            async_dir_cnt[2] % min_copy_records
        ) == 0, f"TID={tid}:\n\t{async_dir_cnt}"
        assert (
            async_dir_cnt[3] % min_copy_records
        ) == 0, f"TID={tid}:\n\t{async_dir_cnt}"


def test_ancestor_ids(input_data):
    """
    This test ensures that each memcpy can be traced back to
    a hipMemcpyAsync through ancestor IDs
    """
    from rocprofiler_sdk.pytest_utils.dotdict import dotdict

    data = input_data
    sdk_data = data["rocprofiler-sdk-json-tool"]
    buffer_records = sdk_data.buffer_records
    memcopies = buffer_records.memory_copies

    _, hip_op_ids = get_operation(buffer_records, "HIP_RUNTIME_API")
    hip_memcopy_id = get_operation(buffer_records, "HIP_RUNTIME_API", "hipMemcpyAsync")

    # dict with { internal id : record }
    hip_memcopies = groupby_corr_id(buffer_records.hip_api_traces, hip_memcopy_id)

    hsa_records = groupby_corr_id(buffer_records.hsa_api_traces)
    hip_records = groupby_corr_id(buffer_records.hip_api_traces)

    accounted_for_hip_ids = []

    for memcpy in memcopies:
        parent_hsa_call = hsa_records[memcpy.correlation_id.internal]
        parent_hip_call = hip_records[parent_hsa_call.correlation_id.ancestor]

        assert (
            parent_hip_call.thread_id == parent_hsa_call.thread_id
        ), "Expected hsa and hip calls to be on the same thread"
        assert hip_op_ids[parent_hip_call.operation] == "hipMemcpyAsync"
        accounted_for_hip_ids.append(parent_hip_call.correlation_id.internal)

    # Ensure we looked through all HIP entries
    assert (
        set(accounted_for_hip_ids) == set(hip_memcopies.keys()),
        "Expected to account for all HIP memcpy calls through ancestor ID lookup",
    )


def test_retired_correlation_ids(input_data):
    data = input_data
    sdk_data = data["rocprofiler-sdk-json-tool"]
    buffer_records = sdk_data["buffer_records"]
    api_name_info = {}

    def _sort_dict(inp):
        return dict(sorted(inp.items()))

    api_corr_ids = {}
    for titr in test_api_traces:
        for itr in sdk_data["buffer_records"][titr]:
            corr_id = itr["correlation_id"]["internal"]
            name = get_operation_name(buffer_records, itr["kind"], itr["operation"])

            assert corr_id not in api_corr_ids.keys()
            assert name is not None, f"{itr}"

            api_corr_ids[corr_id] = itr
            api_name_info[corr_id] = name

    async_corr_ids = {}
    for titr in ["kernel_dispatch", "memory_copies"]:
        for itr in sdk_data["buffer_records"][titr]:
            corr_id = itr["correlation_id"]["internal"]
            assert corr_id not in async_corr_ids.keys()
            async_corr_ids[corr_id] = itr

    retired_corr_ids = {}
    for itr in sdk_data["buffer_records"]["retired_correlation_ids"]:
        corr_id = itr["internal_correlation_id"]
        assert corr_id not in retired_corr_ids.keys()
        retired_corr_ids[corr_id] = itr

    api_corr_ids = _sort_dict(api_corr_ids)
    async_corr_ids = _sort_dict(async_corr_ids)
    retired_corr_ids = _sort_dict(retired_corr_ids)

    #
    # verify all the correlation ids were retired
    #
    num_api_corr_ids = len(api_corr_ids.keys())
    num_retired_corr_ids = len(retired_corr_ids.keys())

    missing_retired_corr_ids = [
        itr for itr in api_corr_ids.keys() if itr not in retired_corr_ids.keys()
    ]
    # log in case of failure
    sys.stderr.flush()
    for itr in missing_retired_corr_ids:
        name = api_name_info[itr]
        info = api_corr_ids[itr]
        sys.stderr.write(f"- unretired corr id: {itr} :: {name} :: {info}\n")
    sys.stderr.flush()

    assert (
        num_api_corr_ids == num_retired_corr_ids
    ), f"correlation ids not retired:\n\t{missing_retired_corr_ids}"

    #
    #   verify the retirement timestamp is >= the end timestamp of the records
    #
    for cid, itr in api_corr_ids.items():
        assert cid in retired_corr_ids.keys()
        retired_ts = retired_corr_ids[cid]["timestamp"]
        end_ts = itr["end_timestamp"]
        name = api_name_info[cid]
        assert (
            retired_ts - end_ts
        ) >= 0, f"\n\tcorr: {cid}\n\tname: {name}\n\tdata: {itr}"

    # allow the retired timestamp to be 10 usec earlier than async end timestamp
    # since the async timestamps undergo conversion from the GPU clock domain to
    # the CPU clock domain. 10 microseconds was arbitrarily chosen to be an
    # acceptable amount of inaccuracy -- in an ideal world, retired_ts should
    # always be >= end_ts
    usec = 1000
    supported_fuzzing = 10 * usec

    for cid, itr in async_corr_ids.items():
        assert cid in retired_corr_ids.keys()
        retired_ts = retired_corr_ids[cid]["timestamp"]
        end_ts = itr["end_timestamp"]
        name = api_name_info[cid]
        assert (
            retired_ts - end_ts
        ) >= -supported_fuzzing, f"\n\tcorr: {cid}\n\tname: {name}\n\tdata: {itr}"


if __name__ == "__main__":
    exit_code = pytest.main(["-x", __file__] + sys.argv[1:])
    sys.exit(exit_code)



################################################################################
# Copyright (c) 2018 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
################################################################################

import sys, os, re
from sqlitedb import SQLiteDB

pinned = ["hipMallocHost", "hipHostMalloc", "hipHostAlloc"]
ondevice = [
    "hipMalloc",
    "hipMallocPitch",
    "hipMallocArray",
    "hipMalloc3DArray",
    "hsa_amd_memory_pool_allocate",
]

mm_table_descr = [
    [
        "BeginNs",
        "EndNs",
        "pid",
        "tid",
        "Name",
        "Direction",
        "SrcType",
        "DstType",
        "Size",
        "Async",
    ],
    {
        "BeginNs": "INTEGER",
        "EndNs": "INTEGER",
        "pid": "INTEGER",
        "tid": "INTEGER",
        "Name": "TEXT",
        "Direction": "TEXT",
        "SrcType": "TEXT",
        "DstType": "TEXT",
        "Size": "INTEGER",
        "Async": "TEXT",
    },
]


def fatal(msg):
    sys.stderr.write(sys.argv[0] + ": " + msg + "\n")
    sys.exit(1)


DELIM = ","


# Mem copy manager class
class MemManager:
    def __init__(self, db, indir):
        self.db = db
        self.allocations = {}
        self.hsa_agent_types = {}
        self.memcopies = {}
        self.filename = ""
        self.fd = 0
        self.parse_hsa_handles(indir + "/" + "hsa_handles.txt")

    def __del__(self):
        if self.fd != 0:
            self.fd.close()

    # Parsing the mapping of HSA agent and memory pool handles
    def parse_hsa_handles(self, infile):
        if os.path.exists(infile):
            inp = open(infile, "r")
            cpu_agent_ptrn = re.compile(r"(0x[0-9a-fA-F]+) agent cpu")
            gpu_agent_ptrn = re.compile(r"(0x[0-9a-fA-F]+) agent gpu")
            for line in inp.readlines():
                m_cpu = cpu_agent_ptrn.match(line)
                if m_cpu:
                    self.hsa_agent_types[str(int(m_cpu.group(1), 16))] = 0  # "cpu"
                m_gpu = gpu_agent_ptrn.match(line)
                if m_gpu:
                    self.hsa_agent_types[str(int(m_gpu.group(1), 16))] = 1  # "gpu"
            inp.close()

    # register alloc and memcpy API calls
    # ['BeginNs', 'EndNs', 'pid', 'tid', 'Name', 'args', 'Index', 'Data'],
    def register_api(self, rec_vals):
        res = ""
        record_name = rec_vals[4]  # 'Name'
        record_args = rec_vals[5]  # 'args'
        malloc_ptrn = re.compile(r"hip.*Malloc|hsa_amd_memory_pool_allocate")
        mcopy_ptrn = re.compile(r"hipMemcpy|hsa_amd_memory_async_copy")

        if malloc_ptrn.match(record_name):
            self.add_allocation(record_name, record_args)
        elif mcopy_ptrn.match(record_name):
            res = self.add_memcpy(rec_vals)

        return res

    # register memcpy asynchronous copy
    # ['BeginNs', 'EndNs', 'Name', 'pid', 'tid', 'Index', ...
    def register_copy(self, rec_vals):
        data = ""
        event = rec_vals[2]  # 'Name'
        procid = rec_vals[3]  # 'pid'
        recordid = rec_vals[5]  # 'Index'
        # query syncronous memcopy API record
        key = (recordid, procid, 0)
        if key in self.memcopies:
            data = self.memcopies[key]

        # query asyncronous memcopy API record
        key = (recordid, procid, 1)
        if key in self.memcopies:
            if data != "":
                fatal("register_copy: corrupted record sync/async")
            async_copy_start_time = rec_vals[0]
            async_copy_end_time = rec_vals[1]

            tid = rec_vals[4]
            copy_line_header = (
                str(async_copy_start_time)
                + DELIM
                + str(async_copy_end_time)
                + DELIM
                + str(procid)
                + DELIM
                + str(tid)
            )
            copy_line_footer = "Async=" + str(1)
            data = copy_line_header + self.memcopies[key] + copy_line_footer
            self.memcopies[key] = data

        return data

    # register memcpy asynchronous activity
    # rec_vals: ['BeginNs', 'EndNs', 'dev-id', 'queue-id', 'Name', 'pid', 'tid', 'Index', 'Data', ...
    def register_activity(self, rec_vals):
        data = ""
        procid = rec_vals[5]  # 'pid'
        recordid = rec_vals[7]  # 'Index'

        # query syncronous memcopy API record
        key = (recordid, procid, 0)
        if key in self.memcopies:
            data = self.memcopies[key]

        # query asyncronous memcopy API record
        key = (recordid, procid, 1)
        if key in self.memcopies:
            if data != "":
                fatal("register_activity: corrupted record sync/async")

            async_copy_start_time = rec_vals[0]
            async_copy_end_time = rec_vals[1]

            tid = rec_vals[6]
            copy_line_header = (
                str(async_copy_start_time)
                + DELIM
                + str(async_copy_end_time)
                + DELIM
                + str(procid)
                + DELIM
                + str(tid)
            )
            copy_line_footer = "Async=" + str(1)
            data = copy_line_header + self.memcopies[key] + copy_line_footer
            self.memcopies[key] = data

        return data

    # add allocation to map
    def add_allocation(self, event, args):
        choice = 0
        if event == "hipMallocPitch":
            malloc_args_ptrn = re.compile(r"\(ptr\((.*)\) width\((.*)\) height\((.*)\)\)")
            choice = 1
        elif event == "hipMallocArray":
            malloc_args_ptrn = re.compile(
                r"\(array\((.*)\) width\((.*)\) height\((.*)\)\)"
            )
            choice = 1
        elif event == "hipMalloc3DArray":
            malloc_args_ptrn = re.compile(
                r"\(array\((.*)\) width\((.*)\) height\((.*)\) depth\((.*)\)\)"
            )
            choice = 2
        elif event == "hsa_amd_memory_pool_allocate":
            # ({handle=25291264}, 40, 0, 0x7ffc4c7bf1b0)
            malloc_args_ptrn = re.compile(
                r"\({handle=\d+}, (\d+), \d+, (0x[0-9a-fA-F]+)\)"
            )
            choice = 4
        else:
            # (ptr(0x7f3407000000) size(800000000) flags(0))
            malloc_args_ptrn = re.compile(r"\(ptr\((.*)\) size\((.*)\) .*\)")
            choice = 3
        m = malloc_args_ptrn.match(args)
        if m:
            if choice == 4:
                ptr = int(m.group(2), 16)
                size = int(m.group(1))
            elif choice == 3:
                ptr = int(m.group(1), 16)
                size = int(m.group(2))
            elif choice == 1:
                ptr = int(m.group(1), 16)
                size = int(m.group(2)) * int(m.group(3))
            else:
                ptr = int(m.group(1), 16)
                size = int(m.group(2)) * int(m.group(3)) * int(m.group(4))
            self.allocations[ptr] = (size, event)

    # get type of ptr
    def get_ptr_type(self, ptr):
        addr = int(ptr, 16)
        addr_type = "unknown"
        found = 0
        for base, (size, event) in self.allocations.items():
            if addr >= base and addr < base + size:
                found = 1
                break
        if not found:
            addr_type = "pageable"
        elif event in pinned:
            addr_type = "pinned"
        elif event in ondevice:
            addr_type = "device"
        elif ptr in self.hsa_agent_types:
            if self.hsa_agent_types[ptr] == 0:
                addr_type = "pinned"
            elif self.hsa_agent_types[ptr] == 1:
                addr_type = "device"
            else:
                fatal("internal error: ptr(" + ptr + ") cannot be identified")
        else:
            fatal("internal error: ptr(" + ptr + ") cannot be identified")
        return addr_type

    # add memcpy to map
    def add_memcpy(self, recvals):
        recordid = recvals[6]  # same as corrid
        event = recvals[4]
        start_time = recvals[0]  # sync time stamp
        end_time = recvals[1]  # sync time stamp
        args = recvals[5]
        procid = int(recvals[2])  # used to query async entries
        pid = recvals[2]
        tid = recvals[3]

        # hipMemcpy(void* dst, const void* src, size_t sizeBytes, hipMemcpyKind kind)
        hip_memcpy_ptrn = re.compile(
            r"\(\s*dst\((.*)\) src\((.*)\) sizeBytes\((\d+)\).*\)"
        )
        # hipMemcpy2D(void* dst, size_t dpitch, const void* src, size_t spitch, size_t width,
        #                   size_t height, hipMemcpyKind kind);
        hip_memcpy_ptrn2 = re.compile(
            r"\(\s*dst\((.*)\) .* src\((.*)\) .* width\((\d+)\) height\((\d+)\).*\)"
        )
        # hipMemcpyToArray(hipArray* dst, size_t wOffset, size_t hOffset, const void* src,
        #                        size_t count, hipMemcpyKind kind);
        hip_memcpy_ptrn3 = re.compile(
            r"\(\s*dst\((.*)\) .* src\((.*)\) count\((\d+)\).*\)"
        )
        # hipMemcpyToSymbol(const void* symbolName, const void* src, size_t sizeBytes,
        #    size_t offset = 0, hipMemcpyKind kind)
        hip_memcpy_ptrn4 = re.compile(
            r"\(\s*symbol\((.*)\) src\((.*)\) sizeBytes\((\d+)\).*\)"
        )
        # memcopy with kind argument
        hip_memcpy_ptrn_kind = re.compile(r".* kind\((\d+)\)\s*.*")
        # hsa_amd_memory_async_copy(void* dst, hsa_agent_t dst_agent, const void* src,
        #                          hsa_agent_t src_agent, size_t size,
        #                          uint32_t num_dep_signals,
        #                          const hsa_signal_t* dep_signals,
        #                          hsa_signal_t completion_signal);
        # "(0x7f8ab6600000, 27064880, 0x7f8b16000000, 27059968, 800000000, 0, 0, 140240759809536) = 0"
        # hsa_memcpy_ptrn_prev used to support format transition and will be cleaned up later.
        hsa_memcpy_ptrn_prev = re.compile(
            r"\((0x[0-9a-fA-F]+), (\d+), (0x[0-9a-fA-F]+), (\d+), (\d+), .*\) = \d"
        )
        # "(0x7fd83bc00000, {handle=16124864}, 0x7fd89b600000, {handle=16119808}, 800000000, 0, 0, {handle=140573877724672}) = 0"
        hsa_memcpy_ptrn = re.compile(
            r"\((0x[0-9a-fA-F]+), {handle=(\d+)}, (0x[0-9a-fA-F]+), {handle=(\d+)}, (\d+), .*\) = \d"
        )
        #    "(0x7f9125cfe7b0, 0x7f9125cfe784, 0x7f9125cfe790, 0x7f9125cfe784, 0x7f9125cfe778, {handle=94324038652880}, 1, 0, 0, {handle=140261380710784}) = 0"
        #    dst, dst_offset, src, src_offset, range, copy_agent, dir, num_dep_signals, dep_signals, completion_signal
        hsa_memcpy_ptrn2 = re.compile(
            r"\((0x[0-9a-fA-F]+), 0x[0-9a-fA-F]+, (0x[0-9a-fA-F]+), 0x[0-9a-fA-F]+, 0x[0-9a-fA-F]+, {z=(\d+), y=(\d+), x=(\d+)}, {handle=(\d+)}, .*\) = \d"
        )
        # aysnc memcopy
        async_event_ptrn = re.compile(r"Async|async")
        m_basic_hip = hip_memcpy_ptrn.match(args)
        m_basic_hsa3 = hip_memcpy_ptrn4.match(args)
        m_basic_hsa_prev = hsa_memcpy_ptrn_prev.match(args)
        m_basic_hsa = hsa_memcpy_ptrn.match(args)
        m_basic_hsa2 = hsa_memcpy_ptrn2.match(args)
        is_hip = True if not (m_basic_hsa_prev or m_basic_hsa or m_basic_hsa2) else False
        m_2d = hip_memcpy_ptrn2.match(args)
        m_array = hip_memcpy_ptrn3.match(args)
        is_async = 1 if async_event_ptrn.search(event) else 0
        copy_line = ""
        size = 0
        dstptr_type = "unknown"
        srcptr_type = "unknown"
        direction = "unknown"
        kind_switcher = {
            "0": "HtoH",
            "1": "HtoD",
            "2": "DtoH",
            "3": "DtoD",
            "4": "auto",
        }

        condition_matched = False
        if m_basic_hip:
            dstptr = m_basic_hip.group(1)
            dstptr_type = self.get_ptr_type(dstptr)
            srcptr = m_basic_hip.group(2)
            srcptr_type = self.get_ptr_type(srcptr)
            size = int(m_basic_hip.group(3))
            condition_matched = True

        if m_basic_hsa_prev:
            dstptr = m_basic_hsa_prev.group(1)
            dst_agent_ptr = m_basic_hsa_prev.group(2)
            dstptr_type = self.get_ptr_type(dst_agent_ptr)
            srcptr = m_basic_hsa_prev.group(3)
            src_agent_ptr = m_basic_hsa_prev.group(4)
            srcptr_type = self.get_ptr_type(src_agent_ptr)
            size = int(m_basic_hsa_prev.group(5))
            condition_matched = True

        if m_basic_hsa:
            dstptr = m_basic_hsa.group(1)
            dst_agent_ptr = m_basic_hsa.group(2)
            dstptr_type = self.get_ptr_type(dst_agent_ptr)
            srcptr = m_basic_hsa.group(3)
            src_agent_ptr = m_basic_hsa.group(4)
            srcptr_type = self.get_ptr_type(src_agent_ptr)
            size = int(m_basic_hsa.group(5))
            condition_matched = True

        if m_basic_hsa2:
            dstptr = m_basic_hsa2.group(1)
            dst_agent_ptr = m_basic_hsa2.group(6)
            dstptr_type = self.get_ptr_type(dst_agent_ptr)
            srcptr = m_basic_hsa2.group(2)
            src_agent_ptr = m_basic_hsa2.group(6)
            srcptr_type = self.get_ptr_type(src_agent_ptr)
            z = int(m_basic_hsa2.group(3))
            y = int(m_basic_hsa2.group(4))
            x = int(m_basic_hsa2.group(5))
            size = x * y * z
            condition_matched = True

        if m_basic_hsa3:
            dstptr = m_basic_hsa3.group(1)
            dstptr_type = self.get_ptr_type(dstptr)
            srcptr = m_basic_hsa3.group(2)
            srcptr_type = self.get_ptr_type(srcptr)
            size = int(m_basic_hsa3.group(3))
            condition_matched = True

        if m_array:
            dstptr = m_array.group(1)
            dstptr_type = self.get_ptr_type(dstptr)
            srcptr = m_array.group(2)
            srcptr_type = self.get_ptr_type(srcptr)
            size = m_array.group(3)
            condition_matched = True

        if m_2d:
            dstptr = m_2d.group(1)
            dstptr_type = self.get_ptr_type(dstptr)
            srcptr = m_2d.group(2)
            srcptr_type = self.get_ptr_type(srcptr)
            size = int(m_2d.group(3)) * int(m_2d.group(4))
            condition_matched = True

        if not condition_matched:
            fatal('Memcpy args "' + args + '" cannot be identified')

        if not is_async:
            start_time = recvals[0]  # sync time stamp
            end_time = recvals[1]  # sync time stamp
            duration = int(end_time) - int(start_time)

        evt_switcher = {
            "hipMemcpyDtoD": "DtoD",
            "hipMemcpyDtoDAsync": "DtoD",
            "hipMemcpyDtoH": "DtoH",
            "hipMemcpyDtoHAsync": "DtoH",
            "hipMemcpyHtoD": "HtoD",
            "hipMemcpyHtoDAsync": "HtoD",
        }

        if is_hip:
            m = hip_memcpy_ptrn_kind.match(args)
            if m:
                direction = kind_switcher.get(m.group(1), "unknown")
            else:
                direction = evt_switcher.get(event, "unknown")
        else:
            if (
                dst_agent_ptr in self.hsa_agent_types
                and src_agent_ptr in self.hsa_agent_types
            ):
                if self.hsa_agent_types[src_agent_ptr] == 1:
                    direction = "D"
                elif self.hsa_agent_types[src_agent_ptr] == 0:
                    direction = "H"
                if direction != "unknown":
                    direction += "to"
                if self.hsa_agent_types[dst_agent_ptr] == 1:
                    direction += "D"
                elif self.hsa_agent_types[dst_agent_ptr] == 0:
                    direction += "H"

        copy_line_header = ""
        copy_line_footer = ""
        copy_line_header = (
            str(start_time) + DELIM + str(end_time) + DELIM + str(pid) + DELIM + str(tid)
        )
        copy_line_footer = "Async=" + str(is_async)

        copy_line = (
            copy_line_header
            + DELIM
            + event
            + DELIM
            + "Direction="
            + direction
            + DELIM
            + "SrcType="
            + srcptr_type
            + DELIM
            + "DstType="
            + dstptr_type
            + DELIM
            + "Size="
            + str(size)
            + DELIM
            + copy_line_footer
        )

        self.memcopies[(recordid, procid, is_async)] = copy_line
        return copy_line

    def dump_data(self, table_name, file_name):
        # To create memcopy info table in DB
        print("File '" + file_name + "' is generating")
        table_handle = self.db.add_table(table_name, mm_table_descr)

        fld_ptrn = re.compile(r"(.*)=(.*)")
        for key, record in self.memcopies.items():
            rec_vals_array = []
            for rec in record.split(DELIM):
                fld_ptrnm = fld_ptrn.match(rec)
                if fld_ptrnm:
                    rec_vals_array.append(fld_ptrnm.group(2))
                else:
                    rec_vals_array.append(rec)
            self.db.insert_entry(table_handle, rec_vals_array)

        # To dump the memcopy info table as CSV
        self.db.dump_csv(table_name, file_name)



################################################################################
# Copyright (c) 2018 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
################################################################################

import os, sys, re


# gen_params() takes a text file like the output of rocminfo cmd and parses it into a map {key,value}
# where key is the param and value is the value of this param
# for example: Threadmodel : "posix"
# it also processes encompasing sections to generate a full param name such as (section names separated by '_'):
#     "Agent2_PoolInfo_ISAInfo_ISA1_WorkgroupMaxSizeperDimension_x": "1024(0x400)",
def gen_params(txtfile):
    fields = {}
    counter = 0
    parent_field = ""
    nbr_indent = 0
    nbr_indent_prev = 0
    check_for_dims = False
    with open(txtfile) as fp:
        for line in fp:
            me = re.match(r"\*\*\* Done \*\*\*", line)  # Marks the end of cmd
            if me:
                parent_field = ""
                nbr_indent = 0
                nbr_indent_prev = 0
                check_for_dims = False
                continue
            mv = re.match(
                r"HCC clang version\s+(.*)", line
            )  # outlier: only line with a version number and no ':', special case
            if mv:
                key = "HCCclangversion"
                val = mv.group(1)
                counter = counter + 1
                fields[(counter, key)] = val
                continue
            # Variable 'check_for_dims' is True for text like this:
            # Workgroup Max Size per Dimension:
            #     x                        1024(0x400)
            #     y                        1024(0x400)
            #     z                        1024(0x400)
            if check_for_dims == True:
                mc = re.match(r"\s*([x|y|z])\s+(.*)", line)
                if mc:
                    key_sav = mc.group(1)
                    if parent_field != "":
                        key = parent_field + "." + mc.group(1)
                    else:
                        key = mc.group(1)
                    val = re.sub(r"\s+", "", mc.group(2))
                    counter = counter + 1
                    fields[(counter, key)] = val
                    if key_sav == "z":
                        check_for_dims = False
            nbr_indent_prev = nbr_indent
            mi = re.search(r"^(\s+)\w+.*", line)
            md = re.search(":", line)
            if mi:
                nbr_indent = int(len(mi.group(1)) / 2)  # indentation cnt
            else:
                if not md:
                    tmp = re.sub(r"\s+", "", line)
                    if tmp.isalnum():
                        parent_field = tmp

            if nbr_indent < nbr_indent_prev:
                go_back_parent = nbr_indent_prev - nbr_indent
                for i in range(go_back_parent):  # decrease as many levels up as needed
                    pos = parent_field.rfind(".")
                    if pos != -1:
                        parent_field = parent_field[:pos]
            # Process lines such as :
            # Segment:                 GLOBAL; FLAGS: KERNARG, FINE GRAINED
            # Size:                    131897644(0x7dc992c) KB
            for lin in line.split(";"):
                lin = re.sub(r"\s+", "", lin)
                m = re.match(r"(.*):(.*)", lin)
                if m:
                    key, val = m.group(1), m.group(2)
                    if parent_field != "":
                        key = parent_field + "." + key
                    if val == "":
                        mk = re.match(r".*Dimension", key)
                        if mk:  # expect x,y,z on next 3 lines
                            check_for_dims = True
                        parent_field = key
                    else:
                        counter = counter + 1
                        fields[(counter, key)] = val
                else:
                    if nbr_indent != nbr_indent_prev and not check_for_dims:
                        parent_field = parent_field + "." + lin.replace(":", "")

    return fields



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import pytest
import json

from collections import defaultdict


# helper function
def node_exists(name, data, min_len=1):
    assert name in data
    assert data[name] is not None
    if isinstance(data[name], (list, tuple, dict, set)):
        assert len(data[name]) >= min_len


def get_operation(record, kind_name, op_name=None):
    for idx, itr in enumerate(record["strings"]["buffer_records"]):
        if kind_name == itr["kind"]:
            if op_name is None:
                return idx, itr["operations"]
            else:
                for oidx, oname in enumerate(itr["operations"]):
                    if op_name == oname:
                        return oidx
    return None


def test_memory_allocation(json_data):
    data = json_data["rocprofiler-sdk-tool"]
    buffer_records = data["buffer_records"]

    memory_allocation_data = buffer_records["memory_allocation"]

    _, bf_op_names = get_operation(data, "MEMORY_ALLOCATION")

    assert len(bf_op_names) == 5

    # Op values:
    UNKNOWN_OP = 0
    HSA_MEMORY_ALLOCATE_OP = 1
    HSA_AMD_VMEM_HANDLE_CREATE_OP = 2
    HSA_MEMORY_FREE_OP = 3
    HSA_AMD_VMEM_HANDLE_RELEASE = 4
    TOTAL_MEM_OPS = 5

    ALLOCATE_OPS = (HSA_MEMORY_ALLOCATE_OP, HSA_AMD_VMEM_HANDLE_CREATE_OP)
    FREE_OPS = (HSA_MEMORY_FREE_OP, HSA_AMD_VMEM_HANDLE_RELEASE)

    valid_agent_ids = set()
    for row in data["agents"]:
        if "id" in row and "handle" in row.id:
            valid_agent_ids.add(row.id.handle)

    memory_alloc_cnt = dict(
        [
            (idx, {"agent": set(), "starting_addr": set(), "size": set(), "count": 0})
            for idx in range(1, TOTAL_MEM_OPS)
        ]
    )

    # check buffering data
    for node in memory_allocation_data:
        assert "size" in node
        assert "kind" in node
        assert "operation" in node
        assert "correlation_id" in node
        assert "end_timestamp" in node
        assert "start_timestamp" in node
        assert "thread_id" in node

        assert "agent_id" in node
        assert "address" in node
        assert "allocation_size" in node

        assert node.size > 0
        assert len(node.address) > 0
        assert node.thread_id > 0

        # Check that op ID is valid
        op_id = node.operation
        assert op_id > UNKNOWN_OP and op_id < TOTAL_MEM_OPS
        # Summarize info in dict
        memory_alloc_cnt[op_id]["count"] += 1
        memory_alloc_cnt[op_id]["starting_addr"].add(node.address)
        memory_alloc_cnt[op_id]["size"].add(node.allocation_size)
        memory_alloc_cnt[op_id]["agent"].add(node.agent_id.handle)

        # Check if agent is valid
        if op_id in ALLOCATE_OPS:
            assert node.agent_id.handle in valid_agent_ids
        else:
            assert node.agent_id.handle == 0  # free ops record agent id as null

        assert node.start_timestamp > 0
        assert node.end_timestamp > 0
        assert node.start_timestamp < node.end_timestamp

        assert data.strings.buffer_records[node.kind].kind == "MEMORY_ALLOCATION"
        assert (
            data.strings.buffer_records[node.kind].operations[node.operation]
            in bf_op_names
        )

    # In the memory allocation test which generates this file
    # 6 hsa_memory_allocation calls with 1024 bytes were called
    # and 9 hsa_amd_memory_pool_allocations with 2048 bytes
    # were called
    assert memory_alloc_cnt[HSA_MEMORY_ALLOCATE_OP]["count"] == 15
    assert memory_alloc_cnt[HSA_MEMORY_FREE_OP]["count"] == 15

    # Check that allocation operations have corresponding free operations
    assert len(memory_alloc_cnt[HSA_MEMORY_ALLOCATE_OP]["starting_addr"]) == len(
        memory_alloc_cnt[HSA_MEMORY_FREE_OP]["starting_addr"]
    )
    for starting_addr in memory_alloc_cnt[HSA_MEMORY_ALLOCATE_OP]["starting_addr"]:
        assert starting_addr in memory_alloc_cnt[HSA_MEMORY_FREE_OP]["starting_addr"]

    # Confirm validation sizes are valid
    assert len(memory_alloc_cnt[HSA_MEMORY_ALLOCATE_OP]["size"]) == 2
    assert (
        len(memory_alloc_cnt[HSA_MEMORY_FREE_OP]["size"]) == 1
    )  # size for free ops are 0
    assert 1024 in memory_alloc_cnt[HSA_MEMORY_ALLOCATE_OP]["size"]
    assert 2048 in memory_alloc_cnt[HSA_MEMORY_ALLOCATE_OP]["size"]

    # Confirm that two agents were used
    assert len(memory_alloc_cnt[HSA_MEMORY_ALLOCATE_OP]["agent"]) == 2


def test_otf2_data(otf2_data, json_data):
    import rocprofiler_sdk.tests.rocprofv3 as rocprofv3

    rocprofv3.test_otf2_data(otf2_data, json_data, ("memory_allocation",))


def test_csv_data(csv_data):
    assert len(csv_data) > 0, "Expected non-empty csv data"

    ALLOCATION_OPS = (
        "MEMORY_ALLOCATION_ALLOCATE",
        "ROCPROFILER_MEMORY_ALLOCATION_VMEM_ALLOCATE",
    )
    FREE_OPS = (
        "ROCPROFILER_MEMORY_ALLOCATION_FREE",
        "ROCPROFILER_MEMORY_ALLOCATION_VMEM_FREE",
    )

    memory_allocation_info = dict(
        {
            "agents": set(),
            "allocation_size": defaultdict(int),
            "address": defaultdict(lambda: defaultdict(int)),
        }
    )

    for row in csv_data:
        assert (
            "Kind" in row
        ), "'Kind' was not present in csv data for memory-allocation-trace"
        assert (
            "Operation" in row
        ), "'Operation' was not present in csv data for memory-allocation-trace"
        assert (
            "Agent_Id" in row
        ), "'Agent_Id' was not present in csv data for memory-allocation-trace"
        assert (
            "Allocation_Size" in row
        ), "'Allocation_Size' was not present in csv data for memory-allocation-trace"
        assert (
            "Address" in row
        ), "'Address' was not present in csv data for memory-allocation-trace"
        assert (
            "Correlation_Id" in row
        ), "'Correlation_Id' was not present in csv data for memory-allocation-trace"
        assert (
            "Start_Timestamp" in row
        ), "'Start_Timestamp' was not present in csv data for memory-allocation-trace"
        assert (
            "End_Timestamp" in row
        ), "'End_Timestamp' was not present in csv data for memory-allocation-trace"

        assert row["Kind"] == "MEMORY_ALLOCATION"
        assert row["Operation"] in (
            "MEMORY_ALLOCATION_ALLOCATE",
            "MEMORY_ALLOCATION_FREE",
        )
        assert int(row["Correlation_Id"]) > 0

        # Check if agent ID is assigned to correct row
        if row["Operation"] in ALLOCATION_OPS:
            assert row["Agent_Id"].split(" ")[0] == "Agent"
            memory_allocation_info["agents"].add(row["Agent_Id"])
        else:
            assert row["Agent_Id"] == ""  # free ops have no agent

        # Confirm allocation size is valid
        if row["Operation"] in ALLOCATION_OPS:
            assert int(row["Allocation_Size"]) in (
                1024,
                2048,
            )  # Test allocates 1024 and 2048 bytes only
            memory_allocation_info["allocation_size"][int(row["Allocation_Size"])] += 1
        else:
            row["Allocation_Size"] == 0  # Free ops record allocation size as 0

        # Confirm address is valid
        assert row["Address"][:2] == "0x"
        address = int(row["Address"], 16)
        if row["Operation"] in ALLOCATION_OPS:
            memory_allocation_info["address"][address]["allocation"] += 1
        else:
            memory_allocation_info["address"][address]["free"] += 1
        # Timestamp sanity check
        assert int(row["Start_Timestamp"]) > 0
        assert int(row["End_Timestamp"]) > 0
        assert int(row["Start_Timestamp"]) < int(row["End_Timestamp"])
    # Test uses 2 agents with 6 allocations of 1024 bytes and 9 allocations of 2048 bytes
    assert len(memory_allocation_info["agents"]) == 2
    assert memory_allocation_info["allocation_size"][1024] == 6
    assert memory_allocation_info["allocation_size"][2048] == 9
    for address in memory_allocation_info["address"].values():
        assert (
            address["allocation"] == address["free"]
        )  # Free should have corresponding allocate


if __name__ == "__main__":
    exit_code = pytest.main(["-x", __file__] + sys.argv[1:])
    sys.exit(exit_code)



# MIT License
#
# Copyright (c) 2023-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from __future__ import absolute_import

import numpy as np
import pandas as pd


def find_wavefront_size(agents_json):
    """
    Find the wavefront size from the agents JSON data.

    The function returns wave front size of the GPU agent 0.
    """
    gpu_agents = list(filter(lambda agent: agent["type"] == 2, agents_json))
    assert len(gpu_agents) > 0, "No GPU agents found"
    first_gpu_agent = gpu_agents[0]
    wavefront_size = first_gpu_agent["wave_front_size"]
    return wavefront_size


def validate_json_exec_mask_manipulation(
    data_json, pc_sampling_method="host_trap", all_sampled=False
):
    """
    The testing function assumes that all kernels run on the first GPU agent
    """
    wave_size = find_wavefront_size(data_json["agents"])
    unique_kernels_num = wave_size + 1

    # Although functional programming might look more elegant,
    # I was trying to avoid multiple iteration over the list of samples.
    # Thus, I decided to use procedural programming instead.
    # Although, it would be more elegant to wrap some of the checks in dedicated functions,
    # I noticed that it can introduce significant overhead, so I decided to inline those checks.

    # the function assume homogenous system
    agents = data_json["agents"]
    gpu_agents = list(filter(lambda agent: agent["type"] == 2, agents))
    # There should be at least one GPU agent
    assert len(gpu_agents) > 0
    first_gpu_agent = gpu_agents[0]
    num_xcc = first_gpu_agent["num_xcc"]
    max_waves_per_simd = first_gpu_agent["max_waves_per_simd"]
    # For GFX9, this represents the number of SIMDs per CU.
    # For GFX10+, this represents the number of SIMDs per WGP.
    simd_per_cu = 4

    gfx_target_version = first_gpu_agent["gfx_target_version"]
    gfx_ip_major = gfx_target_version // 10000

    instructions = data_json["strings"]["pc_sample_instructions"]
    comments = data_json["strings"]["pc_sample_comments"]

    # how many hex digits we have to represent a single execution mask
    exec_mask_hex_digit_width = wave_size // 4

    # execution mask where even SIMD lanes are active
    # correspond to the v_rcp_f64 instructions of the last kernel
    even_simds_active_exec_mask = np.uint64(int("5" * exec_mask_hex_digit_width, 16))
    # start and end source code lines of the v_rcp_f64 instructions of the last kernel
    v_rcp_f64_start_line_num, v_rcp_f64_end_line_num = 288, 387
    # execution mask where even SIMD lanes are active
    # correspond to the v_rcp_f64 instructions of the last kernel
    odd_simds_active_exec_mask = np.uint64(int("A" * exec_mask_hex_digit_width, 16))
    # start and end source code lines of the v_rcp_f32 0 instructions of the last kernel
    v_rcp_f32_start_line_num, v_rcp_f32_end_line_num = 391, 490

    # sampled wave_ids of the last kernel
    last_kernel_sampled_wave_in_grp = set()
    # sampled source lines of the last kernel matching v_rcp_f64 instructions
    last_kernel_v_rcp_64_sampled_source_line_set = set()
    # sampled source lines of the last kernel matching v_rcp_f64 instructions
    last_kernel_v_rcp_f32_sampled_source_line_set = set()
    # sampled correlation IDs
    sampled_cids_set = set()
    # pairs of sampled SIMD ids and waveslot IDs
    sampled_simd_waveslots_pairs = set()
    # sampled chiplets
    sampled_chiplets = set()
    # sample VMIDs
    sampled_vmids = set()
    # TODO: Similar reason for introducing stochastic_assert inside the csv.py.
    # When asserting certain conditions related to exec_masks for all samples,
    # we observe some failures.
    # This usually happens because some small number of samples (e.g., 1-10 out of 100k)
    # do not satisfy the condition. This is either a regression in the ROCr 2nd level trap
    # handler (as sometimes execution mask or correlation ID mismatches), or
    # just stochastic nature of the sampling (meaning our checks are too strict).
    # To relax checks, we introduce an assertion that will allow some small number
    # of samples to disobey the condition.
    # This is a temporary solution until we find the root cause of the issue.

    failing_exec_mask_checks_samples_num = 0
    # We noticed failing samples in:
    # 1. kernels 1-64
    # 2. kernel 65 even SIMD lanes
    # 3. kernel 64 odd SIMD lanes
    # The number of failing samples is less than 10 per category.
    max_number_of_failing_records = 60

    for sample in data_json["buffer_records"][f"pc_sample_{pc_sampling_method}"]:
        record = sample["record"]
        cid = record["corr_id"]["internal"]

        # pull information from hw_id
        hw_id = record["hw_id"]
        sampled_chiplets.add(hw_id["chiplet"])
        sampled_simd_waveslots_pairs.add((hw_id["simd_id"], hw_id["wave_id"]))
        sampled_vmids.add(hw_id["vm_id"])

        # Checks specific for all samples

        # cids must be non-negative numbers
        assert cid >= 0

        inst_index = sample["inst_index"]

        # Since we're not enabling any kind of API tracing, the internal correlation id should
        # be equal to the dispatch_id
        assert cid == record["dispatch_id"]

        if cid == 0:
            # Samples originates either from a blit kernel or self-modifying code.
            # Thus, code object is uknown, as well as the instruction.
            assert record["pc"]["code_object_id"] == 0
            assert inst_index == -1
        else:
            # Update set of sampled cids
            sampled_cids_set.add(cid)

            # All samples with non-zero correlation ID should pass the following checks
            # code object is know, so as the instruction
            assert record["pc"]["code_object_id"] != 0
            assert inst_index != -1

            wgid = record["wrkgrp_id"]
            # check corrdinates of the workgroup
            assert wgid["x"] >= 0 and wgid["x"] <= 1023
            # FIXME: Navi4x wgid is currently broken
            # assert wgid["y"] == 0
            # assert wgid["z"] == 0

            wave_in_grp = record["wave_in_grp"]
            exec_mask = record["exec_mask"]

            if cid < unique_kernels_num:
                # checks specific for samples from first 64 kernels
                assert wave_in_grp == 0
                # inline if possible
                # validate_json_exec_mask_based_on_cid(sample.record)

                # The function assumes that each kernel launches 1024 blocks.
                # Each block contains number of threads that matches correlation ID of the kernel.
                # The exec mask of a sample should contain number of ones equal to
                # the correlation ID of the kernel during which execution the sample was generated.
                # assert bin(exec_mask).count("1") == cid
                if bin(exec_mask).count("1") != cid:
                    failing_exec_mask_checks_samples_num += 1

                # TODO: Comment out the following code if it causes spurious fails.
                # The more conservative constraint based on the experience follows.
                # The exec mask of sampled instructions of the kernels respect the following pattern:
                # cid -> exec
                # 1 -> 0b1
                # 2 -> 0b11
                # 3 -> 0b111
                # ...
                # 64 -> 0xffffffffffffffff
                exec_mask_str = "0b" + "1" * cid
                # assert np.uint64(exec_mask) == np.uint64(int(exec_mask_str, 2))
                if np.uint64(exec_mask) != np.uint64(int(exec_mask_str, 2)):
                    failing_exec_mask_checks_samples_num += 1
            else:
                # No more than `unique_kernels_num`` cids
                assert cid == unique_kernels_num
                # Monitor wave_in_group being sampled
                last_kernel_sampled_wave_in_grp.add(wave_in_grp)
                # chekcs specific for samples from the last kernel
                assert wave_in_grp >= 0 and wave_in_grp <= 3

                # validate instruction decoding
                inst = instructions[inst_index]
                comm = comments[inst_index]
                # The instruction comment is isually in the following format:
                # /path/to/source/file.cpp:line_num
                line_num = int(comm.split(":")[-1])
                if inst.startswith("v_rcp_f64"):
                    # even SIMD lanes active
                    # assert np.uint64(exec_mask) == even_simds_active_exec_mask
                    if np.uint64(exec_mask) != even_simds_active_exec_mask:
                        failing_exec_mask_checks_samples_num += 1

                    assert (
                        line_num >= v_rcp_f64_start_line_num
                        and line_num <= v_rcp_f64_end_line_num
                    )
                    last_kernel_v_rcp_64_sampled_source_line_set.add(line_num)
                elif inst.startswith("v_rcp_f32"):
                    # odd SIMD lanes active
                    # assert np.uint64(exec_mask) == odd_simds_active_exec_mask
                    if np.uint64(exec_mask) != odd_simds_active_exec_mask:
                        failing_exec_mask_checks_samples_num += 1

                    assert (
                        line_num >= v_rcp_f32_start_line_num
                        and line_num <= v_rcp_f32_end_line_num
                    )
                    last_kernel_v_rcp_f32_sampled_source_line_set.add(line_num)

    if all_sampled:
        # All cids that belongs to the range [1, 65] should be samples
        assert len(sampled_cids_set) == unique_kernels_num

        # all wave_ids that belongs to the range [0, 3] should be sampled for the last kernel
        assert len(last_kernel_sampled_wave_in_grp) == 4

        # all source lines matches v_rcp_f64 instructions of the last kernel should be sampled
        assert len(last_kernel_v_rcp_64_sampled_source_line_set) == (
            v_rcp_f64_end_line_num - v_rcp_f64_start_line_num + 1
        )
        # all source lines matches v_rcp_f32 instructions of the last kernel should be sampled
        assert len(last_kernel_v_rcp_f32_sampled_source_line_set) == (
            v_rcp_f32_end_line_num - v_rcp_f32_start_line_num + 1
        )

        # all chiplets must be sampled
        assert len(sampled_chiplets) == num_xcc
        # all (simd ID, waveslot ID) pairs must be samples
        assert len(sampled_simd_waveslots_pairs) == simd_per_cu * max_waves_per_simd

    # assert chiplet index
    assert all(map(lambda chiplet: 0 <= chiplet < num_xcc, sampled_chiplets))
    # assert (SIMD ID, waveslot ID) combinations
    assert all(
        map(
            lambda simd_waveslot: (0 <= simd_waveslot[0] < simd_per_cu)
            and (0 <= simd_waveslot[1] < max_waves_per_simd),
            sampled_simd_waveslots_pairs,
        )
    )

    # Apparently, not all dispatches must belong to the same VMID,
    # so I'm temporarily disabling the following check.
    # # all samples should belong to the same VMID
    # assert len(sampled_vmids) == 1

    # assert that the number of failing samples is acceptable
    assert (
        failing_exec_mask_checks_samples_num <= max_number_of_failing_records
    ), "Number of failing samples failing exec_mask check is too high"



###############################################################################
# MIT License
#
# Copyright (c) 2023 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
###############################################################################


from . import libpyroctx
from . import context_decorators

__all__ = [
    "mark",
    "profilerPause",
    "profilerResume",
    "getThreadId",
    "rangePush",
    "rangePop",
    "rangeStart",
    "rangeStop",
    "nameOsThread",
    "nameHipDevice",
    "context_decorators",
    "version_info",
]

version_info = {
    "version": "1.1.0",
    "major": int("1"),
    "minor": int("1"),
    "patch": int("0"),
    "git_revision": "e1a6bc5663304b9c586b3254b8920f2981057804",
    "library_arch": "",
    "system_name": "Linux",
    "system_processor": "x86_64",
    "system_version": "6.19.8-arch1-1",
    "compiler_id": "GNU",
    "compiler_version": "15.2.1",
    "rocm_version": "7.2.1",
}


def mark(msg):
    return libpyroctx.roctxMark(msg) if msg is not None else None


def profilerPause(tid=0):
    return libpyroctx.roctxProfilerPause(tid)


def profilerResume(tid=0):
    return libpyroctx.roctxProfilerResume(tid)


def getThreadId():
    return libpyroctx.roctxGetThreadId()


def rangePush(msg):
    return libpyroctx.roctxRangePush(msg)


def rangePop():
    return libpyroctx.roctxRangePop()


def rangeStart(msg):
    return libpyroctx.roctxRangeStart(msg) if msg is not None else None


def rangeStop(id=0):
    return libpyroctx.roctxRangeStop(id) if id is not None else None


def nameOsThread(name):
    return libpyroctx.roctxNameOsThread(name)


# def nameHsaAgent(name, agent):
#     return libpyroctx.roctxNameHsaAgent(name, agent)


def nameHipDevice(name, device_id=0):
    return libpyroctx.roctxNameHipDevice(name, device_id)


# def nameHipStream(name, stream):
#     return libpyroctx.roctxNameHipStream(name, stream)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import

import itertools
import sys
import pytest
import numpy as np
import pandas as pd


def validate_all_agents_are_sampled(
    input_samples_csv: pd.DataFrame,
    input_kernel_trace_csv: pd.DataFrame,
    input_agent_info_csv: pd.DataFrame,
):
    transpose_kernel_source_line_start = 137
    transpose_kernel_source_line_end = 145

    gfx9_gfx12_agents_df = input_agent_info_csv[
        input_agent_info_csv["Name"].apply(
            lambda name: name == "gfx90a"
            or name.startswith("gfx94")
            or name.startswith("gfx95")
            or name.startswith("gfx12")
        )
    ]

    # Extract samples that originates from know code object it
    samples_df = input_samples_csv[input_samples_csv["Dispatch_Id"] != 0].copy()

    # Determine the agent on which sample was generated
    # Note: Agent_Id is in the following format e.g., "Agent 3",
    # that's why we need a log for extracting integer value of the id.
    # Determine the agent on which sample was generated
    samples_df["Agent_Id"] = (
        samples_df["Dispatch_Id"]
        .map(
            input_kernel_trace_csv.set_index("Dispatch_Id")["Agent_Id"]
            .str.split(" ")
            .str[1]
        )
        .astype(np.uint64)
    )
    sampled_agents = samples_df["Agent_Id"].unique()
    sampled_agents_num = len(sampled_agents)
    # all agents must be sampled
    assert sampled_agents_num == len(gfx9_gfx12_agents_df)

    # separate samples per agents
    grouped_samples_per_agent = samples_df.groupby("Agent_Id")
    for agent_id, agent_samples_df in grouped_samples_per_agent:
        sampled_dispatches = agent_samples_df["Dispatch_Id"].unique()
        # at least 1 sampled dispatch per agent
        assert len(sampled_dispatches) >= 1

    # extract decoded samples that are mapped to the transpose.cpp file
    transpose_samples_df = samples_df[
        samples_df["Instruction_Comment"].apply(
            lambda comment: "transpose-all-agents.cpp" in comment
        )
    ].copy()
    # determine the line number for each sample
    transpose_samples_df["Source_Line_Num"] = transpose_samples_df[
        "Instruction_Comment"
    ].apply(lambda source_line: int(source_line.split(":")[-1]))
    # assert that line belongs to a kernel range
    assert (
        (transpose_samples_df["Source_Line_Num"] >= transpose_kernel_source_line_start)
        & (transpose_samples_df["Source_Line_Num"] <= transpose_kernel_source_line_end)
    ).all()



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import re
import sys
import pytest
import json


class TimeWindow(object):

    def __init__(self, beg, end):
        self.offset = beg
        self.duration = end - beg

    def in_region(self, val):
        return val >= self.offset and val <= (self.offset + self.duration)

    def __repr__(self):
        return f"[{self.offset}:{self.offset+self.duration}]"


def check_traces(data, valid_regions, invalid_regions, corrid_records=None):
    for record in data:
        corr_id = record.correlation_id.internal
        rval = (
            corrid_records[corr_id]
            if corrid_records is not None and corr_id in corrid_records
            else record
        )
        valid = [itr for itr in valid_regions if itr.in_region(rval.start_timestamp)]
        assert (
            len(valid) == 1
        ), f"\nrval:\n\t{rval}\nrecord:\n\t{record}\nnot found in valid regions:\n{valid_regions}"

        invalid = [itr for itr in invalid_regions if itr.in_region(rval.start_timestamp)]
        assert (
            len(invalid) == 0
        ), f"\nrval:\n\t{rval}\nrecord:\n\t{record}\nfound in invalid region(s):\n{invalid}"


def test_collection_period_trace(json_data, collection_period_data):
    # Adding 20 us error margin to handle the time taken for the start/stop context to affect the collection
    time_error_margin = 20 * 1e4
    valid_regions = []
    invalid_regions = []
    for period in collection_period_data:
        _start = None
        _stop = None
        if "start" in period.keys():
            _start = period.start.start - time_error_margin
        if "stop" in period.keys():
            _stop = period.stop.stop + time_error_margin

        if _start and _stop:
            valid_regions.append(TimeWindow(_start, _stop))
        elif "duration" in period.keys():
            valid_regions.append(TimeWindow(period.duration.start, period.duration.stop))
        elif _start and not _stop:
            valid_regions.append(TimeWindow(_start, _start + 10e9))  # add 10 seconds

        if "delay" in period.keys():
            invalid_regions.append(TimeWindow(period.delay.start, period.delay.stop))

    data = json_data["rocprofiler-sdk-tool"]
    corrid_records = {}

    for itr in ["hsa_api", "hip_api", "marker_api", "rccl_api"]:
        grp = data.buffer_records[itr]
        check_traces(grp, valid_regions, invalid_regions)
        for record in grp:
            corrid_records[record.correlation_id.external] = record


def test_perfetto_data(pftrace_data, json_data):
    import rocprofiler_sdk.tests.rocprofv3 as rocprofv3

    rocprofv3.test_perfetto_data(
        pftrace_data, json_data, ("hip", "hsa", "marker", "kernel", "memory_copy")
    )


def test_otf2_data(otf2_data, json_data):
    import rocprofiler_sdk.tests.rocprofv3 as rocprofv3

    rocprofv3.test_otf2_data(
        otf2_data, json_data, ("hip", "hsa", "marker", "kernel", "memory_copy")
    )


if __name__ == "__main__":
    exit_code = pytest.main(["-x", __file__] + sys.argv[1:])
    sys.exit(exit_code)



################################################################################
# Copyright (c) 2018 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
################################################################################

import os
import sys
import re
import subprocess
import bisect
from sqlitedb import SQLiteDB
from mem_manager import MemManager
import dform

mcopy_data_enabled = 0

EXT_PID = 0
COPY_PID = 1
HIP_PID = 2
HSA_PID = 3
OPS_PID = 5
GPU_BASE_PID = 6
NONE_PID = -1

max_gpu_id = -1
START_NS = 0

hsa_activity_found = 0

# dependencies dictionary
dep_dict = {}
kern_dep_list = []
last_hip_api_map = {}
hip_streams = []
from_ids = {}

# stream ID map
stream_counter = 0
stream_id_map = {}


def get_stream_index(stream_id):
    global stream_counter
    stream_ind = 0
    if stream_id.lower() != "nil":
        if not stream_id in stream_id_map:
            stream_counter += 1
            stream_ind = stream_counter
            stream_id_map[stream_id] = stream_ind
        else:
            stream_ind = stream_id_map[stream_id]
    return stream_ind


# patching activity records
def activity_record_patching(
    db, ops_table_name, kernel_found, kernel_name, stream_found, stream_ind, select_expr
):
    if kernel_found != 0:
        db.change_rec_fld(ops_table_name, 'Name = "' + kernel_name + '"', select_expr)
    if stream_found != 0:
        db.change_rec_fld(ops_table_name, "tid = " + str(stream_ind), select_expr)


# global vars
table_descr = [["Index", "KernelName"], {"Index": "INTEGER", "KernelName": "TEXT"}]
var_list = table_descr[0]
var_table = {}
#############################################################


def fatal(msg):
    sys.stderr.write(sys.argv[0] + ": " + msg + "\n")
    sys.exit(1)


dbglog_count = 0


def dbglog(msg):
    global dbglog_count
    dbglog_count += 1
    sys.stderr.write(sys.argv[0] + ": " + msg + "\n")
    fatal("error")


#############################################################

# Dumping sysinfo
sysinfo_begin = 1


def metadata_gen(sysinfo_file, sysinfo_cmd):
    global sysinfo_begin
    if not re.search(r"\.txt$", sysinfo_file):
        raise Exception('wrong output file type: "' + sysinfo_file + '"')
    if sysinfo_begin == 1:
        sysinfo_begin = 0
        with open(sysinfo_file, mode="w") as fd:
            fd.write("")
    with open(sysinfo_file, mode="a") as fd:
        fd.write("CMD: " + sysinfo_cmd + "\n")
    status = subprocess.call(
        sysinfo_cmd + " >> " + sysinfo_file, stderr=subprocess.STDOUT, shell=True
    )
    if status != 0:
        raise Exception('Could not run command: "' + sysinfo_cmd + '"')


# parse results method
def parse_res(infile):
    global max_gpu_id
    if not os.path.isfile(infile):
        return
    inp = open(infile, "r", errors="replace")

    beg_pattern = re.compile(r'^dispatch\[(\d*)\], (.*) kernel-name\("([^"]*)"\)')
    prop_pattern = re.compile(r"([\w-]+)\((\w+)\)")
    ts_pattern = re.compile(r", time\((\d*),(\d*),(\d*),(\d*)\)")
    # var pattern below matches a variable name and a variable value from a one
    # line text in the format of for example "WRITE_SIZE (0.2500000000)" or
    # "GRBM_GUI_ACTIVE (27867)" or "TA_TA_BUSY[0]"
    var_pattern = re.compile(r"^\s*([a-zA-Z0-9_]+(?:\[\d+\])?)\s+\((\d+(?:\.\d+)?)\)")
    pid_pattern = re.compile(r"pid\((\d*)\)")

    dispatch_number = 0
    var_table_pid = 0
    for line in inp.readlines():
        record = line[:-1]

        m = pid_pattern.search(record)
        if m and not os.getenv("ROCP_MERGE_PIDS"):
            var_table_pid = int(m.group(1))

        m = var_pattern.match(record)
        if m:
            if not (var_table_pid, dispatch_number) in var_table:
                fatal("Error: dispatch number not found '" + str(dispatch_number) + "'")
            var = m.group(1)
            val = m.group(2)
            var_table[(var_table_pid, dispatch_number)][var] = val
            if not var in var_list:
                var_list.append(var)

        m = beg_pattern.match(record)
        if m:
            dispatch_number = m.group(1)
            if not (var_table_pid, dispatch_number) in var_table:
                var_table[(var_table_pid, dispatch_number)] = {
                    "Index": dispatch_number,
                    "KernelName": '"' + m.group(3) + '"',
                }

                gpu_id = 0
                queue_id = 0
                disp_pid = 0
                disp_tid = 0

                kernel_properties = m.group(2)
                for prop in kernel_properties.split(", "):
                    m = prop_pattern.match(prop)
                    if m:
                        var = m.group(1)
                        val = m.group(2)
                        var_table[(var_table_pid, dispatch_number)][var] = val
                        if not var in var_list:
                            var_list.append(var)
                        if var == "gpu-id":
                            gpu_id = int(val)
                            if gpu_id > max_gpu_id:
                                max_gpu_id = gpu_id
                        if var == "queue-id":
                            queue_id = int(val)
                        if var == "pid":
                            disp_pid = int(val)
                        if var == "tid":
                            disp_tid = int(val)
                    else:
                        fatal(
                            'wrong kernel property "'
                            + prop
                            + '" in "'
                            + kernel_properties
                            + '"'
                        )
                m = ts_pattern.search(record)
                if m:
                    var_table[(var_table_pid, dispatch_number)]["DispatchNs"] = m.group(1)
                    var_table[(var_table_pid, dispatch_number)]["BeginNs"] = m.group(2)
                    var_table[(var_table_pid, dispatch_number)]["EndNs"] = m.group(3)
                    var_table[(var_table_pid, dispatch_number)]["CompleteNs"] = m.group(4)

                    # filling dependenciws
                    from_ns = int(m.group(1))
                    to_ns = int(m.group(2))
                    from_us = int((from_ns - START_NS) / 1000)
                    to_us = int((to_ns - START_NS) / 1000)

                    kern_dep_list.append((from_ns, disp_pid, disp_tid))

                    gpu_pid = GPU_BASE_PID + int(gpu_id)
                    if not disp_pid in dep_dict:
                        dep_dict[disp_pid] = {}
                    dep_proc = dep_dict[disp_pid]
                    if not gpu_pid in dep_proc:
                        dep_proc[gpu_pid] = {
                            "pid": HSA_PID,
                            "from": [],
                            "to": {},
                            "id": [],
                        }
                    dep_str = dep_proc[gpu_pid]
                    to_id = len(dep_str["from"])
                    dep_str["from"].append((from_us, disp_tid, disp_tid))
                    dep_str["to"][to_id] = to_us
                    ##

    inp.close()


#############################################################


# Comparator to sort a dictionary of tuples. This comparator will convert
# the second element of tuple to an int and return the new tuple. Then
# the dictionary can use the default comparison i.e sort by first element,
# then sort by second element.
def tuple_comparator(tupleElem):
    return tupleElem[0], int(tupleElem[1])


# merge results table
def merge_table():
    global var_list
    keys = sorted(var_table.keys(), key=tuple_comparator)

    fields = set(var_table[keys[0]])
    if "DispatchNs" in fields:
        var_list.append("DispatchNs")
        var_list.append("BeginNs")
        var_list.append("EndNs")
        var_list.append("CompleteNs")
    var_list = [x for x in var_list if x in fields]


#############################################################


# dump CSV results
def dump_csv(file_name):
    global var_list
    keys = sorted(var_table.keys(), key=tuple_comparator)

    with open(file_name, mode="w") as fd:
        fd.write(",".join(var_list) + "\n")
        for pid, ind in keys:
            entry = var_table[(pid, ind)]
            dispatch_number = entry["Index"]
            if ind != dispatch_number:
                fatal("Dispatch #" + ind + " index mismatch (" + dispatch_number + ")\n")
            val_list = [entry[var] if (var in entry) else 'None' for var in var_list]
            fd.write(",".join(val_list) + "\n")

    print("File '" + file_name + "' is generating")


#############################################################


# fill kernels DB
def fill_kernel_db(table_name, db):
    global var_list
    keys = sorted(var_table.keys(), key=tuple_comparator)

    for var in set(var_list).difference(set(table_descr[1])):
        table_descr[1][var] = "INTEGER"
    table_descr[0] = var_list

    table_handle = db.add_table(table_name, table_descr)

    for pid, ind in keys:
        entry = var_table[(pid, ind)]
        dispatch_number = entry["Index"]
        if ind != dispatch_number:
            fatal("Dispatch #" + ind + " index mismatch (" + dispatch_number + ")\n")
        val_list = [entry[var] for var in var_list]
        db.insert_entry(table_handle, val_list)


#############################################################

# Fill Ext DB
ext_table_descr = [
    ["BeginNs", "EndNs", "pid", "tid", "Name", "Index", "__section", "__lane"],
    {
        "BeginNs": "INTEGER",
        "EndNs": "INTEGER",
        "pid": "INTEGER",
        "tid": "INTEGER",
        "Name": "TEXT",
        "Index": "INTEGER",
        "__section": "INTEGER",
        "__lane": "INTEGER",
    },
]


def fill_ext_db(table_name, db, indir, trace_name, api_pid):
    global range_data

    file_name = indir + "/" + trace_name + "_trace.txt"
    # tms pid:tid cid:rid:'.....'
    ptrn_val = re.compile(r'(\d+) (\d+):(\d+) (\d+):(\d+):"(.*)"$')

    range_data = {}
    range_stack = {}
    range_map = {}

    if not os.path.isfile(file_name):
        return 0

    record_id = 0
    table_handle = db.add_table(table_name, ext_table_descr)
    with open(file_name, mode="r", errors="replace") as fd:
        for line in fd.readlines():
            record = line[:-1]
            m = ptrn_val.match(record)
            if m:
                tms = int(m.group(1))
                pid = m.group(2)
                tid = int(m.group(3))
                cid = int(m.group(4))
                rid = int(m.group(5))
                msg = m.group(6)

                rec_vals = []
                if not tid in range_data:
                    range_data[tid] = {}

                if cid != 2:
                    rec_vals.append(tms)
                    rec_vals.append(tms + 1)
                    rec_vals.append(pid)
                    rec_vals.append(tid)
                    rec_vals.append(msg)
                    rec_vals.append(record_id)
                    rec_vals.append(api_pid)    # __section
                    rec_vals.append(tid)        # __lane

                if cid == 1:
                    if not pid in range_stack:
                        range_stack[pid] = {}
                    pid_stack = range_stack[pid]
                    if not tid in pid_stack:
                        pid_stack[tid] = []
                    rec_stack = pid_stack[tid]
                    rec_stack.append(rec_vals)
                    continue

                if cid == 2:
                    if not pid in range_stack:
                        fatal("ROCTX range begin not found, pid(" + pid + ")")
                    pid_stack = range_stack[pid]
                    if not tid in pid_stack:
                        fatal("ROCTX range begin not found, tid(" + tid + ")")
                    rec_stack = pid_stack[tid]
                    rec_vals = rec_stack.pop()
                    rec_vals[1] = tms
                    # record the range's start/stop timestamps, its parent (ranges can be nested), and its message.
                    range_start = rec_vals[0]
                    range_stop = tms
                    range_parent = rec_stack[-1][0] if len(rec_stack) != 0 else 0
                    range_msg = rec_vals[4]
                    range_data[tid][range_start] = (range_stop, range_parent, range_msg)

                # range start
                if cid == 3:
                    range_map[rid] = (tms, msg)
                    continue

                # range stop
                if cid == 4:
                    if rid in range_map:
                        # querying start timestamp if rid exists
                        (tms, msg) = range_map[rid]
                        del range_map[rid]
                    else:
                        fatal("range id(" + str(rid) + ") is not found")
                    rec_vals[0] = tms   # begin timestamp
                    rec_vals[4] = msg   # range message
                    rec_vals[7] = 0     # 0 lane for ranges

                db.insert_entry(table_handle, rec_vals)
                record_id += 1

    return 1


#############################################################
# arguments manipulation routines
def get_field(args, field):
    if args == None:
        return (None, 0)
    ptrn1_field = re.compile(r"^.* " + field + r"\(")
    ptrn2_field = re.compile(r"\) .*$")
    ptrn3_field = re.compile(r"\)\)$")
    (field_name, n) = ptrn1_field.subn("", args, count=1)
    if n != 0:
        (field_name, n) = ptrn2_field.subn("", field_name, count=1)
        if n == 0:
            (field_name, n) = ptrn3_field.subn("", field_name, count=1)
    return (field_name, n)


def set_field(args, field, val):
    return re.subn(
        field + r"\(\w+\)([ \)])", field + "(" + str(val) + ")\\1", args, count=1
    )


hsa_patch_data = {}
ops_patch_data = {}

# Fill API DB
api_table_descr = [
    [
        "BeginNs",
        "EndNs",
        "pid",
        "tid",
        "Name",
        "args",
        "Index",
        "Data",
        "__section",
        "__lane",
    ],
    {
        "BeginNs": "INTEGER",
        "EndNs": "INTEGER",
        "pid": "INTEGER",
        "tid": "INTEGER",
        "Name": "TEXT",
        "args": "TEXT",
        "Index": "INTEGER",
        "Data": "TEXT",
        "__section": "INTEGER",
        "__lane": "INTEGER",
    },
]


# Filling API records DB table
# table_name - created DB table name
# db - DB handle
# indir - input directory
# api_name - traced API name
# api_pid - assigned JSON PID
# dep_pid - PID of dependet domain
# dep_list - list of dependet dospatch events
# dep_filtr - registered dependencies by record ID
def fill_api_db(
    table_name, db, indir, api_name, api_pid, dep_pid, dep_list, dep_filtr, expl_id
):
    global hsa_activity_found
    global memory_manager

    range_start_times = {}
    copy_csv = ""

    # Matches normal API records.
    ptrn_api_record = re.compile(r"(\d+):(\d+) (\d+):(\d+) ([^\(]+)(\(.*)$")
    # Matches records with a function name of "unknown" and no parameters.
    # Capture groups 1-4 should match the same information as in ptrn_api_record.
    # Used to avoid modifying ptrn_api_record regex.
    ptrn_api_record_unknown = re.compile(r"(\d+):(\d+) (\d+):(\d+) (unknown).*$")

    hip_mcopy_ptrn = re.compile(r"hipMemcpy|hipMemset")
    hip_wait_event_ptrn = re.compile(r"WaitEvent")
    hip_sync_event_ptrn = re.compile(r"hipStreamSynchronize")
    hip_sync_dev_event_ptrn = re.compile(r"hipDeviceSynchronize")
    hip_graph_ptrn = re.compile(r"hipGraphLaunch")
    wait_event_ptrn = re.compile(r"WaitEvent|hipStreamSynchronize|hipDeviceSynchronize")
    hip_stream_wait_write_ptrn = re.compile(
        r"hipStreamWaitValue64|hipStreamWriteValue64|hipStreamWaitValue32|hipStreamWriteValue32"
    )
    prop_pattern = re.compile(r"([\w-]+)\((\w+)\)")
    beg_pattern = re.compile(r'^dispatch\[(\d*)\], (.*) kernel-name\("([^"]*)"\)')
    hip_strm_cr_event_ptrn = re.compile(r"hipStreamCreate")
    hsa_mcopy_ptrn = re.compile(r"hsa_amd_memory_async_copy")
    ptrn_fixformat = re.compile(r"(\d+:\d+ \d+:\d+ \w+)\(\s*(.*)\)$")
    ptrn_fixkernel = re.compile(r"\s+kernel=(.*)$")
    ptrn_multi_kernel = re.compile(r"(.*):(\d+)$")
    ptrn_corr_id = re.compile(r"\ :(\d*)$")

    file_name = indir + "/" + api_name + "_api_trace.txt"
    if not os.path.isfile(file_name):
        return 0

    hsa_copy_file_name = indir + "/" + "async_copy_trace.txt"
    hsa_copy_file_name_present = 1 if os.path.isfile(file_name) else 0
    hsa_copy_deps = 1 if (api_pid == HSA_PID and hsa_copy_file_name_present == 1) else 0
    print("hsa_copy_deps: " + str(hsa_copy_deps))

    # parsing an input trace file and creating a DB table
    record_id_dict = {}
    table_handle = db.add_table(table_name, api_table_descr)
    with open(file_name, mode="r", errors="replace") as fd:
        file_lines = fd.readlines()
        total_lines = len(file_lines)
        line_index = 0
        for line in file_lines:
            if (line_index == total_lines - 1) or (line_index % 100 == 0):
                sys.stdout.write(
                    "\rscan "
                    + api_name
                    + " API data "
                    + str(line_index)
                    + ":"
                    + str(total_lines)
                    + " " * 100
                )
            line_index += 1

            record = line[:-1]

            corr_id = 0
            m = ptrn_corr_id.search(record)
            if m:
                corr_id = int(m.group(1))
                record = ptrn_corr_id.sub("", record)

            kernel_arg = ""
            m = ptrn_fixkernel.search(record)
            if m:
                kernel_arg = "kernel(" + m.group(1) + ") "
                record = ptrn_fixkernel.sub("", record)

            mfixformat = ptrn_fixformat.match(record)
            if mfixformat:  # replace '=' in args with parentheses
                reformated_args = (
                    kernel_arg
                    + mfixformat.group(2)
                    .replace("=", "(")
                    .replace(",", ")")
                    .replace("\\", "\\\\")
                    .replace('"', '\\"')
                    + ")"
                )
                record = mfixformat.group(1) + "( " + reformated_args + ")"

            m = ptrn_api_record.match(record)
            if not m:
                m = ptrn_api_record_unknown.match(record)
                if not m:
                    fatal(api_name + " bad record: '" + record + "'")

            rec_vals = []
            rec_len = len(api_table_descr[0]) - 3
            for ind in range(1, rec_len):
                try:
                    rec_vals.append(m.group(ind))
                except IndexError:
                    rec_vals.append(None)
            proc_id = int(rec_vals[2])
            thread_id = int(rec_vals[3])
            record_name = rec_vals[4]
            # record_args is optional, and may be None if an unknown record is found.
            record_args = rec_vals[5]

            # incrementing per-process record id/correlation id
            if not proc_id in record_id_dict:
                record_id_dict[proc_id] = 0
            record_id_dict[proc_id] += 1
            record_id = record_id_dict[proc_id]

            # setting correlationid to record id if correlation id is not defined
            if corr_id == 0:
                corr_id = record_id

            rec_vals.append(corr_id)
            # extracting/converting stream id
            (stream_id, stream_found) = get_field(record_args, "stream")
            if stream_found:
                stream_id = get_stream_index(stream_id)
                (rec_vals[5], found) = set_field(record_args, "stream", stream_id)
                if found == 0:
                    fatal(
                        'set_field() failed for "stream", args: "' + record_args + '"'
                    )
            else:
                (stream_id, stream_found) = get_field(record_args, "hStream")
                if stream_found:
                    stream_id = get_stream_index(stream_id)
                    (rec_vals[5], found) = set_field(
                        record_args, "hStream", stream_id
                    )
                    if found == 0:
                        fatal(
                            'set_field() failed for "stream", args: "'
                            + record_args
                            + '"'
                        )
                else:
                    stream_id = 0

            if hip_strm_cr_event_ptrn.match(record_name):
                hip_streams.append(stream_id)

            if hip_sync_event_ptrn.match(record_name):
                if (proc_id, stream_id) in last_hip_api_map:
                    (last_hip_api_corr_id, last_hip_api_from_pid) = last_hip_api_map[
                        (proc_id, stream_id)
                    ][-1]
                    sync_api_beg_us = int((int(rec_vals[0]) - START_NS) / 1000)
                    if not proc_id in dep_dict:
                        dep_dict[proc_id] = {}
                    if HIP_PID not in dep_dict[proc_id]:
                        dep_dict[proc_id][HIP_PID] = {
                            "pid": last_hip_api_from_pid,
                            "from": [],
                            "to": {},
                            "id": [],
                        }
                    dep_dict[proc_id][HIP_PID]["from"].append(
                        (-1, stream_id, thread_id)
                    )
                    dep_dict[proc_id][HIP_PID]["id"].append(last_hip_api_corr_id)
                    dep_dict[proc_id][HIP_PID]["to"][
                        last_hip_api_corr_id
                    ] = sync_api_beg_us
                    from_ids[(last_hip_api_corr_id, proc_id)] = (
                        len(dep_dict[proc_id][HIP_PID]["from"]) - 1
                    )

            m = beg_pattern.match(record)
            gpu_id = 0
            if m:
                kernel_properties = m.group(2)
                for prop in kernel_properties.split(", "):
                    m = prop_pattern.match(prop)
                    if m:
                        val = m.group(2)
                        var = m.group(1)
                        if var == "gpu-id":
                            gpu_id = int(val)

            if hsa_mcopy_ptrn.match(record_name) or hip_mcopy_ptrn.match(record_name):
                ops_section_id = COPY_PID
            else:
                ops_section_id = GPU_BASE_PID + int(gpu_id)

            if (proc_id, stream_id) not in last_hip_api_map:
                last_hip_api_map[(proc_id, stream_id)] = []
            last_hip_api_map[(proc_id, stream_id)].append((corr_id, ops_section_id))

            # asyncronous opeartion API found
            op_found = 0
            mcopy_found = 0

            # extract kernel name string
            (kernel_str, kernel_found) = get_field(record_args, "kernel")
            if kernel_found == 0:
                kernel_str = ""
            else:
                op_found = 1

            if hip_mcopy_ptrn.match(record_name):
                mcopy_found = 1
                op_found = 1

            # HIP Graph API
            if hip_graph_ptrn.search(record_name):
                op_found = 1

            # HIP WaitEvent API
            if wait_event_ptrn.search(record_name):
                op_found = 1

            if hip_stream_wait_write_ptrn.search(record_name):
                op_found = 1

            # HSA memcopy API
            if hsa_mcopy_ptrn.match(record_name):
                mcopy_found = 1
                op_found = 1

                stream_id = thread_id
                hsa_patch_data[(corr_id, proc_id)] = thread_id

            if op_found:
                roctx_msg = ""

                if not thread_id in range_start_times:
                    range_start_times[thread_id] = (
                        sorted(range_data[thread_id].keys())
                        if thread_id in range_data
                        else []
                    )
                start_times = range_start_times[thread_id]

                index = bisect.bisect_right(start_times, int(rec_vals[0]))
                if index > 0:
                    # We found the range that is closest to this operation. Iterate the
                    # range stack this range is part of until we find a range that entirely
                    # contains the operation.
                    range_start = start_times[index - 1]
                    while range_start != 0:
                        (range_end, range_start, msg) = range_data[thread_id][
                            range_start
                        ]
                        if int(rec_vals[1]) < range_end:
                            # This range contains the operation.
                            roctx_msg = msg
                            break

                ops_patch_data[(corr_id, proc_id)] = (
                    thread_id,
                    stream_id,
                    kernel_str,
                    roctx_msg,
                )

            if op_found:
                op_found = 0
                beg_ns = int(rec_vals[0])
                end_ns = int(rec_vals[1])
                dur_us = int((end_ns - beg_ns) / 1000)
                from_us = int((beg_ns - START_NS) / 1000) + dur_us / 2
                if api_pid == HIP_PID or hsa_copy_deps == 1:
                    if not proc_id in dep_dict:
                        dep_dict[proc_id] = {}
                    dep_proc = dep_dict[proc_id]
                    if not dep_pid in dep_proc:
                        if api_pid == "HIP_PID":
                            dep_proc[dep_pid] = {"pid": api_pid, "from": [], "id": []}
                        else:
                            dep_proc[dep_pid] = {
                                "pid": api_pid,
                                "from": [],
                                "id": [],
                                "to": {},
                            }
                    dep_str = dep_proc[dep_pid]
                    dep_str["from"].append((from_us, stream_id, thread_id))
                    if expl_id:
                        dep_str["id"].append(corr_id)

            # memcopy registering
            api_data = (
                memory_manager.register_api(rec_vals) if mcopy_data_enabled else ""
            )
            rec_vals.append(api_data)

            # setting section and lane
            rec_vals.append(api_pid)    # __section
            rec_vals.append(thread_id)  # __lane

            # inserting an API record to DB
            db.insert_entry(table_handle, rec_vals)

    # inserting of dispatch events correlated to the dependent dispatches
    for from_ns, proc_id, thread_id in dep_list:
        if not proc_id in record_id_dict:
            record_id_dict[proc_id] = 0
        record_id_dict[proc_id] += 1
        corr_id = record_id_dict[proc_id]
        db.insert_entry(
            table_handle,
            [
                from_ns,
                from_ns,
                proc_id,
                thread_id,
                "hsa_dispatch",
                "",
                corr_id,
                "",
                api_pid,
                thread_id,
            ],
        )

    # generating memcopy CSV
    if copy_csv != "":
        file_name = os.environ["PWD"] + "/results_mcopy.csv"
        with open(file_name, mode="w") as fd:
            print("File '" + file_name + "' is generating")
            fd.write(copy_csv)

    return 1


#############################################################

# fill COPY DB
copy_table_descr = [
    ["BeginNs", "EndNs", "Name", "pid", "tid", "Index", "Data", "__section", "__lane"],
    {
        "Index": "INTEGER",
        "Name": "TEXT",
        "args": "TEXT",
        "BeginNs": "INTEGER",
        "EndNs": "INTEGER",
        "pid": "INTEGER",
        "tid": "INTEGER",
        "Data": "TEXT",
        "__section": "INTEGER",
        "__lane": "INTEGER",
    },
]


def fill_copy_db(table_name, db, indir):
    sect_id = COPY_PID
    file_name = indir + "/" + "async_copy_trace.txt"
    ptrn_val = re.compile(r"^(\d+):(\d+) (async-copy):(\d+):(\d+)$")

    if not os.path.isfile(file_name):
        return 0

    table_handle = db.add_table(table_name, copy_table_descr)
    with open(file_name, mode="r", errors="replace") as fd:
        for line in fd.readlines():
            record = line[:-1]
            m = ptrn_val.match(record)
            if not m:
                fatal("bad async-copy entry '" + record + "'")
            else:
                rec_vals = []
                for ind in range(1, 4):
                    rec_vals.append(m.group(ind))
                corr_id = int(m.group(4))
                proc_id = int(m.group(5))

                # querying tid value
                if (corr_id, proc_id) in hsa_patch_data:
                    thread_id = hsa_patch_data[(corr_id, proc_id)]
                else:
                    thread_id = -1

                # completing record
                rec_vals.append(proc_id)    # tid
                rec_vals.append(thread_id)  # tid
                rec_vals.append(corr_id)    # Index

                # registering memcopy information
                activity_data = (
                    memory_manager.register_copy(rec_vals) if mcopy_data_enabled else ""
                )
                rec_vals.append(activity_data)

                # appending straem ID and section ID
                rec_vals.append(COPY_PID)   # __section
                rec_vals.append(thread_id)  # __lane

                # inserting DB activity entry
                db.insert_entry(table_handle, rec_vals)

                # filling dependencies
                to_ns = int(rec_vals[0])
                to_us = int((to_ns - START_NS) / 1000)

                if thread_id != -1:
                    # if not proc_id in dep_dict: dep_dict[proc_id] = {}
                    dep_proc = dep_dict[proc_id]
                    # if not pid in dep_proc: dep_proc[pid] = { 'pid': HSA_PID, 'from': [], 'to': {}, 'id': [] }
                    dep_str = dep_proc[sect_id]
                    dep_str["to"][corr_id] = to_us
                    dep_str["id"].append(corr_id)

    return 1


#############################################################

# fill HCC ops DB
ops_table_descr = [
    [
        "BeginNs",
        "EndNs",
        "dev-id",
        "queue-id",
        "Name",
        "pid",
        "tid",
        "roctx-range",
        "stream-id",
        "Index",
        "Data",
        "__section",
        "__lane",
    ],
    {
        "Index": "INTEGER",
        "Name": "TEXT",
        "args": "TEXT",
        "BeginNs": "INTEGER",
        "EndNs": "INTEGER",
        "dev-id": "INTEGER",
        "queue-id": "INTEGER",
        "pid": "INTEGER",
        "tid": "INTEGER",
        "roctx-range": "TEXT",
        "Data": "TEXT",
        "stream-id": "INTEGER",
        "__section": "INTEGER",
        "__lane": "INTEGER",
    },
]


def fill_ops_db(kernel_table_name, mcopy_table_name, db, indir):
    global max_gpu_id
    file_name = indir + "/" + "hcc_ops_trace.txt"
    ptrn_val = re.compile(r"(\d+):(\d+) (\d+):(\d+) (.*)$")
    ptrn_id = re.compile(r"^([^:]+):(\d+):(\d+)$")
    ptrn_mcopy = re.compile(r"(Memcpy|Copy|Fill)")
    ptrn_barrier = re.compile(r"Marker")

    if not os.path.isfile(file_name):
        return {}

    filtr = {}

    kernel_table_handle = db.add_table(kernel_table_name, ops_table_descr)
    mcopy_table_handle = db.add_table(mcopy_table_name, ops_table_descr)
    with open(file_name, mode="r", errors="replace") as fd:
        file_lines = fd.readlines()
        total_lines = len(file_lines)
        line_index = 0
        for line in file_lines:
            if (line_index == total_lines - 1) or (line_index % 100 == 0):
                sys.stdout.write(
                    "\rscan ops data "
                    + str(line_index)
                    + ":"
                    + str(total_lines)
                    + " " * 100
                )
            line_index += 1

            record = line[:-1]
            m = ptrn_val.match(record)
            if m:
                # parsing trace record
                rec_vals = []
                for ind in range(1, 6):
                    rec_vals.append(m.group(ind))
                label = rec_vals[4]  # record name
                m = ptrn_id.match(label)
                if not m:
                    fatal("bad hcc ops entry '" + record + "'")
                name = m.group(1)
                corr_id = int(m.group(2))
                proc_id = int(m.group(3))

                # checking name for memcopy pattern
                is_barrier = 0
                if ptrn_mcopy.search(name):
                    rec_table_name = mcopy_table_name
                    table_handle = mcopy_table_handle
                    sect_id = COPY_PID
                else:
                    rec_table_name = kernel_table_name
                    table_handle = kernel_table_handle

                    gpu_id = int(rec_vals[2])
                    if gpu_id > max_gpu_id:
                        max_gpu_id = gpu_id

                    if ptrn_barrier.search(name):
                        name = '"<barrier packet>"'
                        is_barrier = 1
                        sect_id = GPU_BASE_PID + int(gpu_id) + 512
                    else:
                        sect_id = GPU_BASE_PID + int(gpu_id)

                thread_id = 0
                stream_id = 0
                roctx_range = ""
                if (corr_id, proc_id) in ops_patch_data:
                    (thread_id, stream_id, name_patch, roctx_range) = ops_patch_data[
                        (corr_id, proc_id)
                    ]
                    if name_patch != "":
                        name = name_patch
                    if roctx_range == "":
                        roctx_range = name
                else:
                    if is_barrier:
                        continue
                    else:
                        if "ROCP_CTRL_RATE" in os.environ:
                            continue
                        else:
                            fatal(
                                "hcc ops data not found: '"
                                + record
                                + "', "
                                + str(corr_id)
                                + ", "
                                + str(proc_id)
                            )

                # activity record
                rec_vals[4] = name              # Name
                rec_vals.append(proc_id)        # pid
                rec_vals.append(thread_id)      # tid
                rec_vals.append(roctx_range)    # roctx-range
                rec_vals.append(stream_id)      # StreamId
                rec_vals.append(corr_id)        # Index

                # registering memcopy information
                activity_data = (
                    memory_manager.register_activity(rec_vals)
                    if mcopy_data_enabled
                    else ""
                )
                rec_vals.append(activity_data)

                # activity record data for stream ID and sction ID
                rec_vals.append(sect_id)    # __section
                rec_vals.append(stream_id)  # __lane

                # inserting DB activity entry
                db.insert_entry(table_handle, rec_vals)

                # registering a dependency filtr
                filtr[(corr_id, proc_id)] = rec_table_name

                # filling a dependencies
                to_ns = int(rec_vals[0])
                to_us = int((to_ns - START_NS) / 1000)

                end_ns = int(rec_vals[1])
                dur_us = int((end_ns - to_ns) / 1000)

                if (corr_id, proc_id) in from_ids:
                    depid = from_ids[(corr_id, proc_id)]
                    from_val = dep_dict[proc_id][HIP_PID]["from"][depid]
                    print("from_val" + str(from_val))
                    from_val_new = (to_us + dur_us, from_val[1], from_val[2])
                    dep_dict[proc_id][HIP_PID]["from"][depid] = from_val_new

                if not proc_id in dep_dict:
                    dep_dict[proc_id] = {}
                dep_proc = dep_dict[proc_id]
                if not sect_id in dep_proc:
                    dep_proc[sect_id] = {"bsp": OPS_PID, "to": {}}
                dep_str = dep_proc[sect_id]
                dep_str["to"][corr_id] = to_us

            else:
                fatal("hcc ops bad record: '" + record + "'")

    return filtr


#############################################################
# main
if len(sys.argv) < 2:
    fatal("Usage: " + sys.argv[0] + " <output CSV file> <input result files list>")

outfile = sys.argv[1]
infiles = sys.argv[2:]
indir = re.sub(r"\/[^\/]*$", r"", infiles[0])
inext = re.sub(r"\s+$", r"", infiles[0])
inext = re.sub(r"^.*(\.[^\.]+)$", r"\1", inext)

dbfile = ""
csvfile = ""

if "ROCP_JSON_REBASE" in os.environ and os.environ["ROCP_JSON_REBASE"] == 0:
    begin_ts_file = indir + "/begin_ts_file.txt"
    if os.path.isfile(begin_ts_file):
        with open(begin_ts_file, mode="r", errors="replace") as fd:
            ind = 0
            for line in fd.readlines():
                val = int(line)
                if ind == 0 or val < START_NS:
                    START_NS = val
                ind += 1
        print("START timestamp found (" + str(START_NS) + "ns)")

if re.search(r"\.csv$", outfile):
    csvfile = outfile
elif re.search(r"\.db$", outfile):
    dbfile = outfile
    csvfile = re.sub(r"\.db$", ".csv", outfile)
else:
    fatal("Bad output file '" + outfile + "'")

if inext == ".txt":
    for f in infiles:
        parse_res(f)
    if len(var_table) != 0:
        merge_table()

if dbfile == "":
    dump_csv(csvfile)
else:
    statfile = re.sub(r"\.csv$", ".stats.csv", csvfile)
    jsonfile = re.sub(r"\.csv$", ".json", csvfile)

    hsa_statfile = re.sub(r"\.stats\.csv$", r".hsa_stats.csv", statfile)
    hip_statfile = re.sub(r"\.stats\.csv$", r".hip_stats.csv", statfile)
    ops_statfile = statfile
    copy_statfile = re.sub(r"\.stats\.csv$", r".copy_stats.csv", statfile)
    memcopy_info_file = re.sub(r"\.stats\.csv$", r".memcopy_info.csv", statfile)
    sysinfo_file = re.sub(r"\.stats\.csv$", r".sysinfo.txt", statfile)
    metadata_gen(sysinfo_file, "rocminfo")

    with open(dbfile, mode="w") as fd:
        fd.truncate()
    db = SQLiteDB(dbfile)
    memory_manager = MemManager(db, indir)

    ext_trace_found = fill_ext_db("rocTX", db, indir, "roctx", EXT_PID)

    hsa_trace_found = fill_api_db(
        "HSA", db, indir, "hsa", HSA_PID, COPY_PID, kern_dep_list, {}, 0
    )
    hsa_activity_found = fill_copy_db("COPY", db, indir)

    hip_trace_found = fill_api_db("HIP", db, indir, "hip", HIP_PID, OPS_PID, [], {}, 1)
    ops_filtr = fill_ops_db("OPS", "COPY", db, indir)

    fill_kernel_db("KERN", db)

    any_trace_found = ext_trace_found | hsa_trace_found | hip_trace_found
    copy_trace_found = 0
    if hsa_activity_found or len(ops_filtr):
        copy_trace_found = 1

    db.open_json(jsonfile)

    if ext_trace_found:
        db.label_json(EXT_PID, "Markers and Ranges", jsonfile)

    if hip_trace_found:
        db.label_json(HIP_PID, "CPU HIP API", jsonfile)

    if hsa_trace_found:
        db.label_json(HSA_PID, "CPU HSA API", jsonfile)

    db.label_json(COPY_PID, "COPY", jsonfile)

    if max_gpu_id >= 0:
        for ind in range(0, int(max_gpu_id) + 1):
            db.label_json(int(ind) + int(GPU_BASE_PID), "GPU" + str(ind), jsonfile)
            db.label_json(int(ind) + int(GPU_BASE_PID) + 512 , "GPU Barriers" + str(ind), jsonfile)

    if ext_trace_found:
        dform.gen_ext_json_trace(db, "rocTX", START_NS, jsonfile)

    if len(var_table) != 0:
        dform.post_process_data(db, "KERN", csvfile)
        dform.gen_table_bins(db, "KERN", statfile, "KernelName", "DurationNs")
        if "BeginNs" in var_list:
            dform.gen_kernel_json_trace(db, "KERN", GPU_BASE_PID, START_NS, jsonfile)

    if hsa_trace_found:
        dform.post_process_data(db, "HSA")
        dform.gen_table_bins(db, "HSA", hsa_statfile, "Name", "DurationNs")
        dform.gen_api_json_trace(db, "HSA", START_NS, jsonfile)

    if copy_trace_found:
        dform.post_process_data(db, "COPY")
        dform.gen_table_bins(db, "COPY", copy_statfile, "Name", "DurationNs")
        dform.gen_api_json_trace(db, "COPY", START_NS, jsonfile)

    if hip_trace_found:
        dform.post_process_data(db, "HIP")
        dform.gen_table_bins(db, "HIP", hip_statfile, "Name", "DurationNs")
        dform.gen_api_json_trace(db, "HIP", START_NS, jsonfile)

    if ops_filtr:
        dform.post_process_data(db, "OPS")
        dform.gen_table_bins(db, "OPS", ops_statfile, "Name", "DurationNs")
        dform.gen_ops_json_trace(db, "OPS", GPU_BASE_PID, START_NS, jsonfile)

    if any_trace_found:
        dep_id = 0
        for proc_id, dep_proc in dep_dict.items():
            for to_pid, dep_str in dep_proc.items():
                if "bsp" in dep_str:
                    bspid = dep_str["bsp"]
                    base_str = dep_proc[bspid]
                    for v in ("pid", "from", "id"):
                        dep_str[v] = base_str[v]
                    base_str["inv"] = 1

            for to_pid, dep_str in dep_proc.items():
                if "inv" in dep_str:
                    continue
                if not "to" in dep_str:
                    continue

                from_pid = dep_str["pid"]
                from_us_list = dep_str["from"]
                to_us_dict = dep_str["to"]
                corr_id_list = dep_str["id"]

                db.flow_json(
                    dep_id,
                    from_pid,
                    from_us_list,
                    to_pid,
                    to_us_dict,
                    corr_id_list,
                    jsonfile,
                )
                dep_id += len(from_us_list)

    db.metadata_json(jsonfile, sysinfo_file)
    db.close_json(jsonfile)

    if mcopy_data_enabled:
        memory_manager.dump_data("MM", memcopy_info_file)

    db.close()

sys.exit(0)
#############################################################



#!/usr/bin/env python3

# Copyright © Advanced Micro Devices, Inc., or its affiliates.
# SPDX-License-Identifier:  MIT

import sys

if sys.version_info[0] < 3:
    raise Exception("Must be using Python 3")

import ctypes
from ctypes import *
import os

HEADER_OFFSET = 62
HEADER_MASK = 0x3
ID_OFFSET = 34
ID_MASK = (1<<28)-1
OFFSET_MASK = (1<<ID_OFFSET)-1

pluginpath = '../../../lib/rocprofiler/libatt_plugin.so'
filedir = os.path.dirname(os.path.realpath(__file__))
attplugin = CDLL(os.path.join(filedir, pluginpath))

attplugin.getSymbolName.restype = c_char_p
attplugin.getSymbolName.argtypes = [c_uint64]

class instruction_info_t(ctypes.Structure):
    _fields_ = [('inst', c_char_p),
                ('cpp', c_char_p),
                ('size', c_size_t)]

attplugin.getInstructionFromAddr.restype = instruction_info_t
attplugin.getInstructionFromAddr.argtypes = [c_uint64]

attplugin.getInstructionFromID.restype = instruction_info_t
attplugin.getInstructionFromID.argtypes = [c_uint32, c_uint64]

attplugin.addDecoder.restype = c_int
attplugin.addDecoder.argtypes = [c_char_p, c_uint32, c_uint64, c_uint64, c_uint64]

attplugin.removeDecoder.restype = c_int
attplugin.removeDecoder.argtypes = [c_uint32, c_uint64]

def IsRawPC(addr):
    return addr >> HEADER_OFFSET == 0

def getID(addr):
    return (addr >> ID_OFFSET) & ID_MASK

def getOffset(addr):
    return addr & OFFSET_MASK

class CodeobjInstance:
    def __init__(self, gpu_id, line):
        tokens = line.split(' ')
        self.load_base = int(tokens[0], 16)
        self.memsize = int(tokens[1], 16)
        self.att_id = int(tokens[2])
        self.fpath = tokens[3]

        path = self.fpath.encode('utf-8')
        self.error = attplugin.addDecoder(path, self.att_id, self.load_base, self.memsize, gpu_id)
        if self.error != 0:
            print('Warning: Could not open', line)
            raise

    def release(self):
        attplugin.removeDecoder(self.att_id, self.load_base)


class CodeobjService:
    def __init__(self, gpu_id, att_kernel_txt, cfunc):
        cfunc.restype = ctypes.c_int
        cfunc.argtypes = [ctypes.c_char_p, ctypes.c_size_t]

        self.classifier = cfunc
        self.last_instance = None
        self.services = {}
        for line in att_kernel_txt:
            try:
                if 'memory://' == line[0:len('memory://')]:
                    continue
                service = CodeobjInstance(gpu_id, line)
                self.services[service.att_id] = service
            except:
                pass

    def ToRawPC(self, addr):
        if IsRawPC(addr):
            return addr
        return self.services[getID(addr)].load_base + getOffset(addr)

    def release(self):
        for _, instance in self.services.items():
            instance.release()

    def GetInstruction(self, addr):
        if not IsRawPC(addr):
            return self.GetInstructionFromID(getID(addr), getOffset(addr))
        else:
            return self.GetInstructionFromAddr(addr)

    def GetInstructionFromAddr(self, addr):
        info_inst = attplugin.getInstructionFromAddr(addr)
        if info_inst.size == 0 or info_inst.inst is None:
            return None
        inst = info_inst.inst.decode()
        cpp = info_inst.cpp
        if cpp:
            cpp = cpp.decode()

        while len(inst) and (inst[0] == '\t' or inst[0] == ' '):
            inst = inst[1:]
        while len(inst) and (inst[-1] == '\t' or inst[-1] == ' '):
            inst = inst[:-1]

        return (self.classifier(info_inst.inst, len(inst)), inst, cpp, info_inst.size)

    def GetInstructionFromID(self, id, offset):
        info_inst = attplugin.getInstructionFromID(id, offset)
        if info_inst.size == 0 or info_inst.inst is None:
            return None
        inst = info_inst.inst.decode()
        cpp = info_inst.cpp
        if cpp:
            cpp = cpp.decode()
        else:
            cpp = ''

        while len(inst) and (inst[0] == '\t' or inst[0] == ' '):
            inst = inst[1:]
        while len(inst) and (inst[-1] == '\t' or inst[-1] == ' '):
            inst = inst[:-1]

        return (self.classifier(info_inst.inst, len(inst)), inst, cpp, info_inst.size)

    def getSymbolName(self, addr):
        try:
            name = attplugin.getSymbolName(self.ToRawPC(addr))
            if name:
                return name.decode()
            return "Addr #"+hex(self.ToRawPC(addr))
        except:
            return "Addr #"+hex(addr)



#!/usr/bin/env python3
###############################################################################
# MIT License
#
# Copyright (c) 2025 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
###############################################################################

from __future__ import absolute_import

__author__ = "Advanced Micro Devices, Inc."
__copyright__ = "Copyright 2025, Advanced Micro Devices, Inc."
__license__ = "MIT"


def main(argv=None, config=None):
    """Main entry point for the rocpd command line tool.

    Args:
        argv (list, optional): List of command line arguments. Defaults to None.

    """
    import argparse
    from . import csv
    from . import otf2
    from . import output_config
    from . import pftrace
    from . import query
    from . import summary
    from . import time_window
    from . import version_info
    from .importer import RocpdImportData

    convert_examples = """

Example usage:

    Convert 1 database, output perfetto trace
    $ rocpd convert -i db1.db --output-format pftrace

    Convert 2 databases, output perfetto trace to path and filename, reduce time window to omit the first 30%
    $ rocpd convert -i db1.db db2.db --output-format pftrace -d "./output/" -o "twoFileTraces" --start 30% --end 100%

    Convert 6 databases, output CSV and perfetto trace formats
    $ rocpd convert -i db{0..5}.db --output-format csv pftrace -d "~/output_folder/" -o "sixFileTraces"

    Convert 2 databases, output CSV, OTF2, and perfetto trace formats
    $ rocpd convert -i db{3,4}.db --output-format csv otf2 pftrace

"""

    query_examples = """

Example usage:

    Query the first 5 rows of the 'rocpd_info_agents' view and output to console
    $ rocpd query -i db0.db --query "SELECT * FROM rocpd_info_agents LIMIT 5"

    Combine 4 databases and query the first 10 rows of the 'top_kernels' view and output to CSV file
    $ rocpd query -i db{0..3}.db --query "SELECT * FROM top_kernels LIMIT 10" --format csv
"""

    summary_examples = """

Example usage:

    Output all summaries to console and include domain summary for 1 database
    $ rocpd summary -i db1.db --domain-summary

    Aggregate 3 databases and output all summary files and include summary by rank/process ID, to csv file output
    $ rocpd summary -i db{1..3}.db --summary-by-rank --format csv

    Output all summaries to console and exlude all regions to save processing time
    $ rocpd summary -i db0.db --region-categories NONE

    Aggregate 2 databases and output all summary files to HTML, only include HIP and MARKER regions, include domain summary
    $ rocpd summary -i db{0,1}.db --region-categories HIP MARKERS --domain-summary --format html

"""

    # Add the subparsers
    parser = argparse.ArgumentParser(
        prog="rocpd",
        description="Aggregate and/or analyze ROCm Profiling Data (rocpd)",
        allow_abbrev=False,
    )

    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="Print the version information and exit",
    )

    subparsers = parser.add_subparsers(dest="command")
    converter = subparsers.add_parser(
        "convert",
        description="Convert rocPD data into another data format",
        allow_abbrev=False,
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=convert_examples,
    )

    query_reporter = subparsers.add_parser(
        "query",
        description="Generate output on a query",
        allow_abbrev=False,
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=query_examples,
    )

    generate_summary = subparsers.add_parser(
        "summary",
        description="Generate summary views from rocPD data",
        allow_abbrev=False,
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=summary_examples,
    )

    def get_output_type(val):
        return val.lower().replace("perfetto", "pftrace")

    # add required options for each subparser
    converter_required_params = converter.add_argument_group("Required options")
    converter_required_params.add_argument(
        "-i",
        "--input",
        required=True,
        type=output_config.check_file_exists,
        nargs="+",
        help="Input path and filename to one or more database(s)",
    )
    converter_required_params.add_argument(
        "-f",
        "--output-format",
        help="For adding output format (supported formats: csv, pftrace, otf2)",
        nargs="+",
        default=None,
        choices=("csv", "pftrace", "otf2"),
        type=get_output_type,
        required=True,
    )

    query_required_params = query_reporter.add_argument_group("Required options")
    query_required_params.add_argument(
        "-i",
        "--input",
        required=True,
        type=output_config.check_file_exists,
        nargs="+",
        help="Input path and filename to one or more database(s)",
    )

    summary_required_params = generate_summary.add_argument_group("Required options")
    summary_required_params.add_argument(
        "-i",
        "--input",
        required=True,
        type=output_config.check_file_exists,
        nargs="+",
        help="Input path and filename to one or more database(s)",
    )

    # converter: add args from any sub-modules
    valid_out_config_args = output_config.add_args(converter)
    valid_generic_args = output_config.add_generic_args(converter)
    valid_pftrace_args = pftrace.add_args(converter)
    valid_csv_args = csv.add_args(converter)
    valid_otf2_args = otf2.add_args(converter)
    valid_time_window_args = time_window.add_args(converter)

    # query: subparser args
    valid_out_config_args = output_config.add_args(query_reporter)
    valid_query_args = query.add_args(query_reporter)
    valid_time_window_args = time_window.add_args(query_reporter)

    # summary: subparser args
    valid_io_args = summary.add_io_args(generate_summary)
    valid_summary_args = summary.add_args(generate_summary)
    valid_time_window_args = time_window.add_args(generate_summary)

    # parse the command line arguments
    args = parser.parse_args(argv)

    if args.version:
        for key, itr in version_info.items():
            if key in ["major", "minor", "patch"]:
                continue
            print(f"    {key:>16}: {itr}")
        return 0

    # error check the command line arguments, if no subparser command is given, print the help message
    if args.command is None:
        parser.print_help()
        return

    # if the user requested converter, process the conversion
    if args.command == "convert":
        # process the args
        out_cfg_args = output_config.process_args(args, valid_out_config_args)
        generic_out_cfg_args = output_config.process_generic_args(
            args, valid_generic_args
        )
        pftrace_args = pftrace.process_args(args, valid_pftrace_args)
        csv_args = csv.process_args(args, valid_csv_args)
        otf2_args = otf2.process_args(args, valid_otf2_args)
        window_args = time_window.process_args(args, valid_time_window_args)

        # now start processing the data.  Import the data and merge the views
        importData = RocpdImportData(args.input)

        # adjust the time window view of the data
        if window_args is not None:
            time_window.apply_time_window(importData, **window_args)

        all_args = {
            **out_cfg_args,
            **generic_out_cfg_args,
            **pftrace_args,
            **csv_args,
            **otf2_args,
        }
        # setup the config args
        config = (
            output_config.output_config(**all_args)
            if config is None
            else config.update(**all_args)
        )

        # process each requested output format
        format_handlers = {
            "pftrace": pftrace.write_pftrace,
            "csv": csv.write_csv,
            "otf2": otf2.write_otf2,
        }

        for out_format in args.output_format:
            if out_format in format_handlers:
                print(f"Converting database(s) to {out_format} format:")
                format_handlers[out_format](importData, config)
            else:
                print(f"Warning: Unsupported output format '{out_format}'")

    # if the user requested query module, execute the query
    elif args.command == "query":
        # query subparser args
        query_args = query.process_args(args, valid_query_args)
        out_cfg_args = output_config.process_args(args, valid_out_config_args)
        window_args = time_window.process_args(args, valid_time_window_args)

        all_args = {**query_args, **out_cfg_args}

        query.execute(
            args.input,
            args,
            window_args=window_args,
            **all_args,
        )

    # if the user requested a summary, generate the views
    elif args.command == "summary":
        # summary subparser args
        summary_args = summary.process_args(args, valid_summary_args)
        io_args = output_config.process_args(args, valid_io_args)
        window_args = time_window.process_args(args, valid_time_window_args)

        # now start processing the data.  Import the data and merge the views
        importData = RocpdImportData(args.input)

        # adjust the time window view of the data
        if window_args is not None:
            time_window.apply_time_window(importData, **window_args)

        all_args = {**summary_args, **io_args}
        summary.generate_all_summaries(importData, **all_args)

    print("Done. Exiting...")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import os
import pytest
import ctypes


def test_validate_metrics(rocm_path):
    set_library(rocm_path)

    from rocprofv3 import avail

    lib = avail.get_library()

    agent_counters_dict = avail.get_counters()
    for agent, counters in agent_counters_dict.items():
        for counter in counters:
            assert counter.name != ""
            assert counter.description != ""
            if isinstance(counter, avail.derived_counter):
                assert counter.expression != ""
            elif isinstance(counter, avail.basic_counter):
                assert counter.block != ""
            assert counter.dimensions != ""
            if counter.name == "TA_BUSY_min":
                counter.description == "TA block is busy. Min over TA instances."
                counter.expression == "reduce(TA_TA_BUSY,min)"
            if counter.name == "TA_BUSY_min":
                counter.description == "TA block is busy. Min over TA instances."
                counter.expression == "reduce(TA_TA_BUSY,min)"


def test_validate_list_pc_sample_config(rocm_path):
    set_library(rocm_path)

    from rocprofv3 import avail

    lib = avail.get_library()

    pc_sample_configs_dict = avail.get_pc_sample_configs()
    for agent, configs in pc_sample_configs_dict.items():
        for config in configs:
            assert config.method != ""
            assert config.unit != ""
            assert isinstance(config.max_interval, ctypes.c_ulong) == True
            assert isinstance(config.min_interval, ctypes.c_ulong) == True
            if config.method == "stochastic":
                assert config.flags.value == 1
            elif config.method == "host_trap":
                assert config.flags.value == 0
            assert config.max_interval.value >= config.min_interval.value


def test_counter_set(capsys, rocm_path):
    set_library(rocm_path)

    from rocprofv3 import avail

    def get_counter_names(counter_ids):

        counter_names = []
        for counter_id in counter_ids:
            counter = get_counter_info(counter_id)
            if counter.counter_handle == counter_id:
                counter_names.append(counter.name)
        return counter_names

    def get_agent_name(agent_id):
        agent_info_map = get_agent_info_map()
        for agent, info in agent_info_map.items():
            if agent == agent_id:
                return info["name"]

    lib = avail.get_library()
    agent_counters = avail.get_counters()
    pmc_input_1 = []

    for agent, counters in agent_counters.items():
        counter_ids = []
        for counter in counters:
            counter_ids.append(counter.counter_handle)
        input = {agent: [counter_ids[0]]}
        assert avail.check_pmc(input) == True
        with pytest.raises(SystemExit) as excinfo:
            input = {agent: counter_ids}
            avail.check_pmc(input)
            output = capsys.readouterr()
            test = "{} not collected on agent {}".format(
                " ".join(get_counter_names(counter_ids)), get_agent_name(agent)
            )
            assert test == output


def set_library(rocm_path):
    ROCPROFV3_AVAIL_PACKAGE = f"{rocm_path}/lib/python{sys.version_info[0]}/site-packages"
    sys.path.append(ROCPROFV3_AVAIL_PACKAGE)

    from rocprofv3 import avail

    ROCPROF_LIST_AVAIL_TOOL_LIBRARY = (
        f"{rocm_path}/lib/rocprofiler-sdk/librocprofv3-list-avail.so"
    )
    os.environ["ROCPROFILER_METRICS_PATH"] = f"{rocm_path}/share/rocprofiler-sdk"
    avail.loadLibrary.libname = os.environ.get(
        "ROCPROF_LIST_AVAIL_TOOL_LIBRARY", ROCPROF_LIST_AVAIL_TOOL_LIBRARY
    )


if __name__ == "__main__":
    exit_code = pytest.main(["-x", __file__] + sys.argv[1:])
    sys.exit(exit_code)



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import pytest
import numpy as np
import pandas as pd
import re


def unique(lst):
    return list(set(lst))


def validate_json(input_data):
    json_data = input_data["json_data"]
    iteration_list = input_data["iteration_range"]

    data = json_data["rocprofiler-sdk-tool"]
    counter_collection_data = data["callback_records"]["counter_collection"]
    kernel_dispatch_data = data["buffer_records"]["kernel_dispatch"]
    dispatch_ids = {}

    def get_kernel_name(kernel_id):
        return data["kernel_symbols"][kernel_id]["formatted_kernel_name"]

    iteration = 1
    for dispatch in kernel_dispatch_data:
        dispatch_info = dispatch["dispatch_info"]
        kernel_name = get_kernel_name(dispatch_info["kernel_id"])
        if kernel_name == "transpose" and iteration in iteration_list:
            dispatch_ids[dispatch_info["dispatch_id"]] = dispatch_info
        if kernel_name == "transpose":
            iteration = iteration + 1

    for counter in counter_collection_data:
        dispatch_data = counter["dispatch_data"]["dispatch_info"]
        dispatch_id = dispatch_data["dispatch_id"]
        if dispatch_id in dispatch_ids.keys():
            assert dispatch_data == dispatch_ids[dispatch_id]


def test_validate_counter_collection_pass1(input_json_pass1):
    validate_json(input_json_pass1)


def test_validate_counter_collection_pass2(input_json_pass2):
    validate_json(input_json_pass2)


def test_validate_counter_collection_pass3(input_json_pass3):
    validate_json(input_json_pass3)


if __name__ == "__main__":
    exit_code = pytest.main(["-x", __file__] + sys.argv[1:])
    sys.exit(exit_code)
